# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for interaction with adb."""
import json
import os
import subprocess
import time

from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import host_utils

import six

ADB_RETRY_SLEEP = 10
FASTBOOT_TIMEOUT = 10.0
PROPERTY_PATTERN = r"\[(.*)\]: \[(.*)\]\n"
SYSENV_PATTERN = r"(.*)=(.*)\n"
logger = gdm_logger.get_logger()


def enter_fastboot(adb_serial, adb_path=None):
  """Enters fastboot mode by calling 'adb reboot bootloader' for the adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from calling 'adb reboot' or None if call fails with
      non-zero
           return code.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  return _adb_command(("reboot", "bootloader"), adb_serial, adb_path=adb_path)


def exit_fastboot(fastboot_serial,
                  fastboot_path=None,
                  timeout=FASTBOOT_TIMEOUT):
  """Exits fastboot mode by calling 'fastboot reboot' for the fastboot_serial provided.

  Args:
      fastboot_serial (str): Device fastboot serial number.
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot reboot to return

  Raises:
      RuntimeError: if fastboot_path is invalid or fastboot executable was not
                    found by get_fastboot_path.

  Returns:
      str: Output from calling 'fastboot reboot' or None if call fails with
      non-zero
           return code.

  Note:
      If fastboot_path is not provided then path returned by get_fastboot_path
      will be used instead. If fastboot returns a non-zero return code then
      None will be returned.
  """
  if fastboot_path is None:
    fastboot_path = get_fastboot_path()

  if not os.path.exists(fastboot_path):
    raise RuntimeError(
        "The fastboot_path of {} appears to be invalid.".format(fastboot_path))

  try:
    args = ("timeout", str(timeout), fastboot_path, "-s", fastboot_serial,
            "reboot")
    return subprocess.check_output(
        args, stderr=subprocess.STDOUT).decode("utf-8", "replace")
  except subprocess.CalledProcessError:
    return None


def fastboot_unlock_device(fastboot_serial,
                           fastboot_path=None,
                           timeout=FASTBOOT_TIMEOUT):
  """Unlock the device through fastboot.

  Args:
      fastboot_serial (str): Device serial number
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot command to return

  Returns:
      str: response from fastboot command
  """
  return _fastboot_command(("flashing", "unlock"),
                           fastboot_serial=fastboot_serial,
                           fastboot_path=fastboot_path,
                           timeout=timeout)


def fastboot_lock_device(fastboot_serial,
                         fastboot_path=None,
                         timeout=FASTBOOT_TIMEOUT):
  """Lock the device through fastboot.

  Args:
      fastboot_serial (str): Device serial number
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot command to return

  Returns:
      str: response from fastboot command
  """
  return _fastboot_command(("flashing", "lock"),
                           fastboot_serial=fastboot_serial,
                           fastboot_path=fastboot_path,
                           timeout=timeout)


def fastboot_wipe_userdata(fastboot_serial,
                           fastboot_path=None,
                           timeout=FASTBOOT_TIMEOUT):
  """Wipe user data on the device through fastboot.

  Args:
      fastboot_serial (str): Device serial number
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot command to return

  Returns:
      str: response from fastboot command
  """
  return _fastboot_command(
      "-w",
      fastboot_serial=fastboot_serial,
      fastboot_path=fastboot_path,
      timeout=timeout)


def enter_sideload(adb_serial, adb_path=None, auto_reboot=False):
  """Enters sideload mode by calling 'adb reboot sideload' for the adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable.
      auto_reboot (bool): whether to auto reboot after sideload complete.

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from command call.
  """
  if auto_reboot:
    command = ("reboot", "sideload-auto-reboot")
  else:
    command = ("reboot", "sideload")
  return _adb_command(command, adb_serial=adb_serial, adb_path=adb_path)


def is_sideload_mode(adb_serial, adb_path=None):
  """Checks if device is in sideload mode.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb_path executable was not
      found by
                    get_adb_path().

  Returns:
      bool: True if device is in sideload mode. False otherwise.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead.
  """
  return adb_serial in get_sideload_devices(adb_path=adb_path)


def sideload_package(package_path, adb_serial, adb_path=None):
  """Perform "adb sideload <package>" command.

  Args:
      package_path (str): the path of the package to sideload.
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable.

  Returns:
      str: the command output.

  Raises:
      RuntimeError: if package_path is invalid.
  """
  if not os.path.isfile(package_path):
    raise RuntimeError(
        "sideload_package failed: {} is not a file.".format(package_path))
  return _adb_command(("sideload", package_path),
                      adb_serial=adb_serial,
                      adb_path=adb_path)


def get_sideload_devices(adb_path=None):
  """Returns a list of adb devices in sideload mode.

  Args:
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      list: A list of device serial numbers that are in sideload mode.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead.
  """
  try:
    output = _adb_command("devices", adb_path=adb_path)
  except RuntimeError as err:
    logger.warning(repr(err))
    return []
  device_lines = [x for x in output.splitlines() if "\tsideload" in x]
  return [x.split()[0] for x in device_lines]


def get_adb_devices(adb_path=None):
  """Returns a list of available adb devices.

  Args:
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      list: A list of device serial numbers returned by 'adb devices'.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead.
  """
  try:
    output = _adb_command("devices", adb_path=adb_path)
  except RuntimeError as err:
    logger.warning(repr(err))
    return []
  device_lines = [x for x in output.splitlines() if "\tdevice" in x]
  return [x.split()[0].split(":", 1)[0] for x in device_lines]


def get_adb_path(adb_path=None):
  """Returns the correct adb path to use.

  Args:
      adb_path (str): path to "adb" executable.

  Notes: Starts with passed in path, then looks at config, and finally
    system's default adb if available.

  Raises:
      RuntimeError: if no valid adb path could be found

  Returns:
      str: Path to correct adb executable to use.
  """
  if is_valid_path(adb_path):
    return adb_path
  try:
    with open(config.DEFAULT_GDM_CONFIG_FILE) as config_file:
      gdm_config = json.load(config_file)
    adb_path = gdm_config[config.ADB_BIN_PATH_CONFIG]
  except (IOError, KeyError, ValueError):
    pass

  if is_valid_path(adb_path):
    return adb_path
  elif adb_path:
    logger.warning("adb path {!r} stored in {} does not exist."
                   .format(adb_path, config.DEFAULT_GDM_CONFIG_FILE))

  if host_utils.has_command("adb"):
    return host_utils.get_command_path("adb")
  raise RuntimeError("No valid adb path found using 'which adb'")


def is_valid_path(path):
  return path and os.path.exists(path)


def connect(adb_serial, adb_path=None):
  """Connects to device via ADB."""
  resp = _adb_command(["connect", adb_serial], adb_path=None)
  if "unable to connect" in str(resp):
    raise errors.DeviceError(
        f"Unable to connect to device {adb_serial!r} via ADB: {resp}")


def shell(adb_serial, command, adb_path=None, timeout=None, retries=1):
  """Issues a command to the shell of the adb_serial provided.

  Args:
      adb_serial (str): Device serial number
      command (str): command to send
      adb_path (str): optional alternative path to adb executable
      timeout (int): time in seconds to wait for adb process to complete.
      retries (int): number of times to retry adb command.

  Returns:
      str: response from adb command

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead.
  """
  return _adb_command(["shell", command], adb_serial,
                      adb_path=adb_path, timeout=timeout, retries=retries)


def get_fastboot_devices(fastboot_path=None):
  """Returns list of ADB devices in fastboot (bootloader) mode.

  Args:
      fastboot_path (str): optional alternative path to fastboot executable

  Returns:
      list: A list of ADB device serial numbers in fastboot mode.

  Note:
      If fastboot_path is not provided then path returned by get_fastboot_path
      will be used instead.
      If fastboot path invalid, will return empty list.
  """
  try:
    fastboot_path = get_fastboot_path(fastboot_path)
  except RuntimeError as err:
    logger.warning(repr(err))
    return []

  try:
    output = subprocess.check_output((fastboot_path, "devices"),
                                     stderr=subprocess.STDOUT)
    output = output.decode("utf-8", "replace")
    device_lines = [x for x in output.splitlines() if "astboot" in x]
    return [x.split()[0] for x in device_lines]
  except subprocess.CalledProcessError:
    return []


def get_fastboot_path(fastboot_path=None):
  """Returns the fastboot executable path to use.

  Args:
      fastboot_path (str): path to "fastboot" executable.

  Raises:
      RuntimeError: if no valid fastboot executable could be found

  Returns:
      str: Path to correct fastboot executable to use.
  """
  if is_valid_path(fastboot_path):
    return fastboot_path
  if host_utils.has_command("fastboot"):
    return host_utils.get_command_path("fastboot")
  raise RuntimeError("No valid fastboot path found using 'which fastboot'")


def is_adb_mode(adb_serial, adb_path=None):
  """Checks if device is in adb mode.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      bool: True if device is in adb mode. False otherwise.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead.
  """
  return adb_serial in get_adb_devices(adb_path=adb_path)


def is_device_online(adb_serial, adb_path=None, fastboot_path=None):
  """Returns true if the device appears in either 'adb devices' or 'fastboot devices'.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable
      fastboot_path (str): optional alternative path to fastboot executable

  Returns:
      bool: True if device is in adb or fastboot mode. False otherwise.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If fastboot_path is not provided then path returned by
      get_fastboot_path will be used instead.
  """
  return (is_adb_mode(adb_serial, adb_path=adb_path) or
          is_fastboot_mode(adb_serial, fastboot_path=fastboot_path))


def is_fastboot_mode(adb_serial, fastboot_path=None):
  """Checks if device is in fastboot mode.

  Args:
      adb_serial (str): Device serial number.
      fastboot_path (str): optional alternative path to fastboot executable

  Raises:
      RuntimeError: if fastboot_path is invalid or fastboot executable was
                    not found by get_fastboot_path.

  Returns:
      bool: True if device is in fastboot mode. False otherwise.

  Note:
      If fastboot_path is not provided then path returned by get_fastboot_path
      will be used instead.
  """
  return adb_serial in get_fastboot_devices(fastboot_path=fastboot_path)


def pull_from_device(adb_serial, sources, destination_path="./", adb_path=None):
  """Pulls sources from device to destination_path on host for adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      sources (str or list): Path to one or more source files on device to
        copy to host.
      destination_path (str): Path to destination on host computer where file
        should copied to.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path or push failed.
      ValueError: if destination_path directory doesn't exist.

  Returns:
      str: Output from calling 'adb push' or None if call raises an erroro
           return code.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned. If no destination_path is provided the file will be copied to
      the current working directory on the host computer.
  """
  destination_dir = os.path.dirname(destination_path)
  if destination_dir != "." and not os.path.exists(destination_dir):
    raise ValueError(
        "The destination_path directory {} appears to be invalid.".format(
            destination_dir))

  args = ["pull"]
  if isinstance(sources, list):
    for source_path in sources:
      args.append(source_path)
  else:
    args.append(sources)
  args.append(destination_path)
  output, returncode = _adb_command(
      args, adb_serial, adb_path=adb_path, include_return_code=True)
  if returncode != 0:
    raise RuntimeError("Pulling file(s) {} on ADB device {} to {} failed. "
                       "Error: {!r}".format(sources, adb_serial,
                                            destination_path, output))
  return output


def push_to_device(adb_serial, sources, destination_path, adb_path=None):
  """Pushes sources to destination_path on device for adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      sources (str or list): Path to one or more source files on host computer
        to copy to device.
      destination_path (str): Path to destination on device where file should
        copied to.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path or push failed.
      ValueError: if source_path doesn't exist.

  Returns:
      str: Output from calling 'adb push' or None if call raises an erroro
           return code.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  args = ["push"]
  if isinstance(sources, list):
    for source_path in sources:
      args.append(source_path)
      if not os.path.exists(source_path):
        raise ValueError(
            "The source file {} appears to be invalid.".format(source_path))
  else:
    args.append(sources)
    if not os.path.exists(sources):
      raise ValueError(
          "The source file {} appears to be invalid.".format(sources))

  args.append(destination_path)
  output, returncode = _adb_command(
      args, adb_serial, adb_path=adb_path, include_return_code=True)
  if returncode != 0:
    raise RuntimeError("Pushing file(s) {} to {} on ADB device {} failed. "
                       "Error: {!r}".format(sources, destination_path,
                                            adb_serial, output))
  return output


def reboot_device(adb_serial, adb_path=None, retries=1):
  """Calls 'adb reboot' for the adb_serial provided using adb_path.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executables
      retries (int): number of times to retry adb command.

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from calling 'adb reboot'.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  return _adb_command("reboot", adb_serial, adb_path=adb_path, retries=retries)


def root_device(adb_serial, adb_path=None):
  """Calls 'adb root' for the adb_serial provided using adb_path.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from calling 'adb root'.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  return _adb_command("root", adb_serial, adb_path=adb_path)


def verify_user_has_fastboot(device_name):
  """Verifies fastboot available and user is root or in plugdev group.

  Args:
      device_name (str): Device name to use in error output.

  Raises:
      DeviceError: Fastboot is not on computer OR
                   'plugdev' group doesn't exist OR
                   current user is not in the 'plugdev' group.
  """
  if not host_utils.has_command("fastboot"):
    raise errors.DeviceError("Device {} verify user has fastboot failed. "
                             "Fastboot executable is not installed. "
                             "See readme about installing adb (which installs "
                             "fastboot) then su -$USER (or logout and back in) "
                             "to add user to plugdev group".format(device_name))


def _adb_command(command,
                 adb_serial=None,
                 adb_path=None,
                 include_return_code=False,
                 timeout=None,
                 retries=1):
  """Returns the output of the adb command and optionally the return code.

  Args:
      command (str or tuple): ADB command and optionally arguments to execute.
      adb_serial (str): Device serial number
      adb_path (str): optional alternative path to adb executable
      include_return_code (bool): flag indicating return code should also be
        returned.
      timeout (int): time in seconds to wait for adb process to complete.
      retries (int): number of times to retry adb command.

  Raises:
      RuntimeError: if adb_path provided or obtained from get_adb_path is
                    invalid (executable at path doesn't exist).

  Returns:
      str: The ADB command output (including stderr)
      tuple: The ADB command output (including stderr) and return code

  Note:
      The stderr is redirected to stdout so callers should use the return code
      or search the output for known errors if they want to determine if the
      command succeeded or not.
  """
  adb_path = get_adb_path(adb_path)

  if adb_serial is None:
    args = [adb_path]
  else:
    args = [adb_path, "-s", adb_serial]
  if isinstance(command, (str, six.text_type)):
    args.append(command)
  elif isinstance(command, (list, tuple)):
    args.extend(command)
  for i in range(0, retries):
    proc = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
      output, _ = proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
      proc.terminate()
      output, _ = proc.communicate()
    output = output.decode("utf-8", "replace")
    logger.debug("adb command {!r} to {} returned {!r}".format(
        command, adb_serial, output))
    if include_return_code:
      return output, proc.returncode
    if not any(msg in output for msg in ["error: closed", "offline"]):
      return output
    if i < retries - 1:
      logger.info(f"Retrying adb command: {command} in {ADB_RETRY_SLEEP}s")
      time.sleep(ADB_RETRY_SLEEP)
  raise errors.DeviceError(
      f"ADB command failed: {command} with output: {output}")


def _fastboot_command(command,
                      fastboot_serial=None,
                      fastboot_path=None,
                      include_return_code=False,
                      timeout=FASTBOOT_TIMEOUT):
  """Returns the output of the fastboot command and optionally the return code.

  Args:
      command (str or tuple): fastboot command and optionally arguments to
        execute.
      fastboot_serial (str): Device fastboot serial number.
      fastboot_path (str): optional alternative path to fastboot executable
      include_return_code (bool): flag indicating return code should also be
        returned.
      timeout (float): in seconds to wait for fastboot command to return

  Raises:
      RuntimeError: if fastboot_path provided or obtained from
      get_fastboot_path is invalid (executable at path doesn't exist).

  Returns:
      str: The fastboot command output (including stderr)
      tuple: The fastboot command output (including stderr) and return code

  Note:
      The stderr is redirected to stdout so callers should use the return code
      or search the output for known errors if they want to determine if the
      command succeeded or not.
  """
  if fastboot_path is None:
    fastboot_path = get_fastboot_path()
  if not os.path.exists(fastboot_path):
    raise RuntimeError(
        "The fastboot_path of {} appears to be invalid.".format(fastboot_path))

  if fastboot_serial is None:
    args = ["timeout", str(timeout), fastboot_path]
  else:
    args = ["timeout", str(timeout), fastboot_path, "-s", fastboot_serial]

  if isinstance(command, (str, six.text_type)):
    args.append(command)
  elif isinstance(command, (list, tuple)):
    args.extend(command)

  proc = subprocess.Popen(
      args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  output, _ = proc.communicate()
  output = output.decode("utf-8", "replace")
  if include_return_code:
    return output, proc.returncode
  return output


def install_package_on_device(package_path,
                              adb_serial=None,
                              adb_path=None,
                              allow_downgrade=False,
                              allow_test_apk=False,
                              reinstall=False):
  """Installs an apk on a target device.

  Use adb install command to install a package to the system.
  The options are subjected to the adb install command. See the doc.
  https://developer.android.com/studio/command-line/adb#shellcommands

  Args:
      package_path (str): the path to the package on host machine.
      adb_serial (str): the device serial, optional.
      adb_path (str): optional alternative path to adb executable.
      allow_downgrade (bool): allows version code downgrade.
      allow_test_apk (bool): allows test APKs to be installed.
      reinstall (bool): reinstalls an existing app and keeps its data.

  Raises:
      ValueError: when pacakge_path is not valid.
      DeviceError: when installation failed.
  """
  if not os.path.exists(package_path):
    raise ValueError(
        "install_package_on_device received invalid package_path: {}".format(
            package_path))

  flags_map = {"-d": allow_downgrade, "-t": allow_test_apk, "-r": reinstall}
  command_list = ["install"]
  flags = sorted([flag for flag, value in flags_map.items() if value])
  command_list.extend(flags)
  command_list.append(package_path)
  response = _adb_command(
      tuple(command_list), adb_serial=adb_serial, adb_path=adb_path)
  if "Success\n" not in response:
    raise errors.DeviceError(
        "install_package_on_device failed: {}".format(response))


def uninstall_package_on_device(package_name, adb_serial=None, adb_path=None):
  """Uninstall a package on a target device.

  Args:
      package_name (str): the name of the package, e.g.,
        "com.google.android.apps.somepackage.someapp".
      adb_serial (str): the device serial, optional.
      adb_path (str): optional alternative path to adb executable.

  Raises:
      DeviceError: when uninstall failed.
  """
  response = _adb_command(("uninstall", package_name),
                          adb_serial=adb_serial,
                          adb_path=adb_path)
  if "Success\n" not in response:
    raise errors.DeviceError("uninstall_package_on_device failed.")



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_unifi_switch capability."""
import re

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import switch_power_base

logger = gdm_logger.get_logger()

TELNET_COMMANDS = {
    "ENTER_TELNET": {
        "ENTER_CMD": ["telnet localhost", "enable"],
        "ENTER_REGEX": [r"\s+>", r"\s+#"],
        "EXIT_CMD": ["exit", "exit"],
        "EXIT_REGEX": [r"\s+>", "Connection closed by foreign host"]
    },
    "ENTER_TELNET_CONFIG": {
        "ENTER_CMD": ["telnet localhost", "enable", "config"],
        "ENTER_REGEX": [r"\s+>", r"\s+#", r"\(Config\)#"],
        "EXIT_CMD": ["exit", "exit", "exit"],
        "EXIT_REGEX": [r"\s+#", r"\s+>", "Connection closed by foreign host"]
    },
    "ENTER_TELNET_PORT_CONFIG": {
        "ENTER_CMD": ["telnet localhost", "enable", "config"],
        "ENTER_REGEX": [r"\s+>", r"\s+#", r"\(Config\)#"],
        "EXIT_CMD": ["exit", "exit", "exit", "exit"],
        "EXIT_REGEX": [
            r"\(Config\)#", r"\s+#", r"\s+>",
            "Connection closed by foreign host"
        ]
    }
}

OFF = "off"
ON = "on"


class SwitchPowerUnifiSwitch(switch_power_base.SwitchPowerBase):
  """Definition of the switch_power_unifi_switch capability."""
  _REQUIRED_COMMANDS = [
      "GET_PORT_MODE", "ADJUST_PORTS_MODE", "GET_ALL_PORTS_MODE"
  ]
  _REQUIRED_REGEXS = ["GET_PORT_MODE_REGEX", "GET_ALL_PORTS_MODE_REGEX"]

  def __init__(self, device_name, command_dict, regex_dict, total_ports,
               send_and_expect_fn, send_fn):
    """Create an instance of the poe switch power capability.

    Args:
        device_name (str): Name of the device this capability is attached
          to.
        command_dict (dict): A dictionary containing the command used for
          each method.
        regex_dict (dict): A dictionary containing the Regex used to
          validate device states or command success.
        total_ports (int): Total number of device ports.
        send_and_expect_fn (func): send_and_expect method of Switchboard
          capability.
        send_fn (func): send method of Switchboard capability.
    """

    super().__init__(device_name=device_name)
    self._command_dict = command_dict
    self._regex_dict = regex_dict
    self._total_ports = total_ports
    self._send_and_expect_fn = send_and_expect_fn
    self._send_fn = send_fn

    self.validate_required_keys(self._REQUIRED_COMMANDS, self._command_dict,
                                "command_dict")
    self.validate_required_keys(self._REQUIRED_REGEXS, self._regex_dict,
                                "regex_dict")

  @decorators.PersistentProperty
  def supported_modes(self):
    """Get the Poe switch power modes supported by the unifi_switch."""
    return [OFF, ON]

  @decorators.PersistentProperty
  def total_ports(self):
    """Get the number of device ports.

    Returns:
        int: the number of device ports
    """
    return int(self._total_ports)

  @decorators.CapabilityLogDecorator(logger)
  def power_on(self, port):
    """Powers on the specified port.

    Args:
        port (int): device port number
    """
    port = int(port)
    self._validate_port("power_on", port)
    telnet_commands = self._create_port_config_command_and_regex(port)
    logger.debug(f"{self._device_name} Powering on unifi_switch port {port}")
    self._telnet_port_config_send(
        telnet_commands=telnet_commands,
        func=self._send_fn,
        func_args=(self._command_dict["ADJUST_PORTS_MODE"].format(
            mode="auto"),))
    self._verify_mode("power_on", port, ON)

  @decorators.CapabilityLogDecorator(logger)
  def power_off(self, port):
    """Powers off the specified port.

    Args:
        port (int): device port number
    """
    port = int(port)
    self._validate_port("power_off", port)
    telnet_commands = self._create_port_config_command_and_regex(port)
    logger.debug(f"{self._device_name} Powering off unifi_switch port {port}")
    self._telnet_port_config_send(
        telnet_commands=telnet_commands,
        func=self._send_fn,
        func_args=(self._command_dict["ADJUST_PORTS_MODE"].format(
            mode="shutdown"),))
    self._verify_mode("power_off", port, OFF)

  @decorators.CapabilityLogDecorator(logger)
  def set_mode(self, mode, port):
    """Sets the given Ethernet port to the mode specified.

    Args:
        mode (str): Mode to set the specified unifi_switch port. e.g. 'off',
          'on'
        port (int): The port to set.
    Note: Not all devices support 'charge'

    Raises:
      DeviceError: invalid port, or mode.
    """
    self._validate_mode(mode)
    if mode == OFF:
      self.power_off(port)
    else:
      self.power_on(port)

  @decorators.CapabilityLogDecorator(logger)
  def set_all_ports_mode(self, mode):
    """Set all device ports to the specified mode.

    Args:
        mode (str): Mode to set all unifi_switch ports to. Valid modes are
          "on" (auto) or "off (shutdown)".
    """
    self._validate_mode(mode)
    for poe_port in range(1, self._total_ports + 1):
      self.set_mode(mode, poe_port)

  def get_mode(self, port):
    """Get mode of the specified port.

    Args:
        port (int): device port number

    Returns:
        str: mode of specified port, either 'on'(auto), 'passive24V', or
        'off'(shutdown).
    """
    port = int(port)
    self._validate_port("get_mode", port)
    response = self._telnet_config_send(
        func=self._poe_telnet_send_and_expect,
        func_args=([self._command_dict["GET_PORT_MODE"].format(port=port)],
                   [self._regex_dict["GET_PORT_MODE_REGEX"].format(port=port)]),
        func_kwargs={"regex_group": 1})

    if response == "Auto":
      return ON
    elif response == "Shutdown":
      return OFF
    else:
      return response

  def get_all_ports_mode(self):
    """Get mode of all the device ports.

    Returns:
        list: a list of the all the ports mode.
    """
    result = self._telnet_config_send(
        func=self._poe_telnet_send_and_expect,
        func_args=([self._command_dict["GET_ALL_PORTS_MODE"]],
                   [self._regex_dict["GET_ALL_PORTS_MODE_REGEX"]]),
        func_kwargs={"regex_group": 1})
    ports_status = []
    for poe_port in range(1, self._total_ports + 1):
      match = re.search(
          self._regex_dict["GET_PORT_MODE_REGEX"].format(port=poe_port), result)
      if match:
        if match.group(1) == "Auto":
          ports_status.append(ON)
        elif match.group(1) == "Shutdown":
          ports_status.append(OFF)
        else:
          ports_status.append(match.group(1))
      else:
        ports_status.append("")
    return ports_status

  def _create_port_config_command_and_regex(self, poe_port):
    """Create the port config command and regex for the specified poe_port."""
    telnet_commands = TELNET_COMMANDS["ENTER_TELNET_PORT_CONFIG"].copy()
    telnet_commands["ENTER_CMD"].append(fr'interface 0/{poe_port}')
    telnet_commands["ENTER_REGEX"].append(fr'\(Interface 0/{poe_port}\)#')
    return telnet_commands

  def _telnet_port_config_send(self,
                               telnet_commands,
                               func,
                               func_args=(),
                               func_kwargs=None):
    """Send commands at the telnet_port_config menu level."""
    if func_kwargs is None:
      func_kwargs = {}
    self._poe_telnet_send_and_expect(telnet_commands["ENTER_CMD"],
                                     telnet_commands["ENTER_REGEX"])
    func_result = func(*func_args, **func_kwargs)
    self._poe_telnet_send_and_expect(telnet_commands["EXIT_CMD"],
                                     telnet_commands["EXIT_REGEX"])
    return func_result

  def _telnet_config_send(self, func, func_args=(), func_kwargs=None):
    """Send commands at the telnet_config menu level."""
    if func_kwargs is None:
      func_kwargs = {}
    self._poe_telnet_send_and_expect(
        TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["ENTER_CMD"],
        TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["ENTER_REGEX"])
    func_result = func(*func_args, **func_kwargs)
    self._poe_telnet_send_and_expect(
        TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["EXIT_CMD"],
        TELNET_COMMANDS["ENTER_TELNET_CONFIG"]["EXIT_REGEX"])
    return func_result

  def _telnet_send(self, func, func_args=(), func_kwargs=None):
    """Send commands at the telnet menu level."""
    if func_kwargs is None:
      func_kwargs = {}
    self._poe_telnet_send_and_expect(
        TELNET_COMMANDS["ENTER_TELNET"]["ENTER_CMD"],
        TELNET_COMMANDS["ENTER_TELNET"]["ENTER_REGEX"])
    func_result = func(*func_args, **func_kwargs)
    self._poe_telnet_send_and_expect(
        TELNET_COMMANDS["ENTER_TELNET"]["EXIT_CMD"],
        TELNET_COMMANDS["ENTER_TELNET"]["EXIT_REGEX"])
    return func_result

  def _validate_port(self, method_name, port):
    """Ensure port is a valid port number.

    Args:
        method_name(str): name of the method where _validate_port is being
          called
        port (int): device port number

    Raises:
         DeviceError: if input port is not valid
    """
    if port is None or port <= 0 or port > self._total_ports:
      raise errors.DeviceError("Device {} {} failed. "
                               "Port {} is invalid.".format(
                                   self._device_name, method_name, port))

  def _validate_mode(self, mode):
    """Verify mode given resides in the valid mode list.

    Args:
        mode (str): mode e.g. 'off', 'on'

    Raises:
        DeviceError: if mode given is not found in the valid mode list.
    """
    if mode not in list(self.supported_modes):
      raise errors.DeviceError("Mode {} not found in supported modes{}".format(
          mode, self.supported_modes))

  def _verify_mode(self, method_name, port, mode):
    """Ensure port is set to specified mode.

    Args:
        method_name(str): name of the method where _validate_port is being
          called
        port (int): device port number
        mode (str): mode of the specified port. Valid modes are 'auto' or
          'shutdown'.

    Raises:
         DeviceError: if port is not in specified mode
    """
    active_mode = self.get_mode(port)
    if active_mode != mode:
      raise errors.DeviceError(
          f"Device {self._device_name} {method_name} failed. "
          f"Mode is {active_mode} not {mode}.")

  def _poe_telnet_send_and_expect(self,
                                  commands,
                                  loglines,
                                  timeout=30.0,
                                  command_tries=1,
                                  raise_for_timeout=True,
                                  regex_group=0):
    """Helper method to send multiple send_and_expect commands during a telnet session.

    Args:
        commands (list): list of commands to send via send_and_expect.
        loglines (list): list of regex expressions to match with the
          corresponding command.
        timeout (float): timeout of send and expect.
        command_tries (int): number of times to retry the command.
        raise_for_timeout(bool): raise exception in event of timeout of
          send_and_expect.
        regex_group (int): regext match group to return.

    Returns:
        str: match of specified regex group.

    Raises:
         DeviceError: if each command doesn't have a corresponding regex.
    """
    if len(loglines) != len(commands):
      raise errors.DeviceError(
          f"The number of commands {len(commands)} does not equal "
          f"the number of loglines {len(loglines)}. "
          f"The number of commands and loglines must be equal.")
    for cmd_num in range(len(commands)):
      response = self._send_and_expect_fn(
          command=commands[cmd_num],
          pattern_list=[loglines[cmd_num]],
          timeout=timeout,
          command_tries=command_tries,
          raise_for_timeout=raise_for_timeout,
          expect_type="response")

    return response.match.group(regex_group)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""USB port map utility."""
import weakref

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.auxiliary_devices import cambrionix
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_logger()

# The PRODUCT_DEVICE_MAP is a dictionary with 'vendor_id/product_id' as the key
# and the 'description' 'type' in the values dict. The description includes
# the device_type in GDM as applicable.
PRODUCT_DEVICE_MAP = {
    '0403/6015': {
        'description': 'Cambrionix FTDI',
        'device_type': 'cambrionix'
    },
    '2cd9/0021': {
        'description': 'Cambrionix USB3',
        'device_type': 'cambrionix'
    },
}

SINGLE_CONNECTION_LIST = [
]

HUB_DEVICES_COLUMN_HEADER = ('    Issue         Mode    Discovered'
                             '                Configured')
OTHER_DEVICES_COLUMN_HEADER = '    Issue         Mode    Discovered'
OTHER_DEVICES_HEADER = 'Other devices'

# Conflict types that port-map will test for and report.
USB_HUB_CONFLICT = 'usb hub conflict'
USB_PORT_CONFLICT = 'usb port conflict'
MISSING_USB_HUB = 'missing usb hub'
MISSING_USB_PORT = 'missing usb port'
NOT_DISCOVERED_CONFLICT = 'not discovered'
NOT_DETECTED_CONFLICT = 'not detected'


def _initialize_default_device(device):
  """Set values to the default for a device dictionary.

  Args:
      device (dict): The dictionary to be initialized with the default device
        data.
  """
  device['detected_devices'] = []
  device['device_type'] = None
  device['address'] = ''
  device['usb_hub_address'] = ''
  device['usb_hub_port'] = ''
  device['type'] = None
  device['serial_number'] = None
  device['conflict_types'] = []
  device['known_device'] = False
  device['description'] = '-'
  device['ftdi_interface'] = 0


def _identify_device(usb_info_inst):
  """Identify a udev usb device object and gather information about it.

  Args:
      usb_info_inst (UsbInfo): usb_info instance with usb information.

  Returns:
      dict: A dict containing device properties with the following keys:
          'description' (required): description of the device, or
          contents of ID_SERIAL if device not recognized.
          'name' (optional): GDM device name of the device, if recognized.
          'path_id' (required): the first part of the device id path.
          'part_id' (optional): the last part of the device id path.
    """
  product = '{}/{}'.format(usb_info_inst.vendor_id, usb_info_inst.product_id)
  device_map_entry = PRODUCT_DEVICE_MAP.get(product)
  device_data = {}
  _initialize_default_device(device_data)
  if device_map_entry:
    device_data['known_device'] = True
    if 'device_type' in device_map_entry:
      id_suffix = usb_info_inst.serial_number[-4:].lower()
      device_data['name'] = '{}-{}'.format(device_map_entry['device_type'],
                                           id_suffix)

    if 'type' in device_map_entry:
      device_data['type'] = device_map_entry['type']

    device_data['description'] = device_map_entry['description']
  else:
    device_data['description'] = usb_info_inst.serial_number

  device_data['serial_number'] = usb_info_inst.serial_number
  device_data['address'] = usb_info_inst.address
  device_data['usb_hub_address'] = usb_info_inst.usb_hub_address
  device_data['usb_hub_port'] = usb_info_inst.usb_hub_port
  device_data['ftdi_interface'] = usb_info_inst.ftdi_interface
  return device_data


def _get_usb_hub_map(device_info_list):
  """Creates a map of usb hub addresses to device_infos by port.

  Args:
      device_info_list (list): list of known usb_connections dicts.

  Returns:
     dict: map of usb hub addresses to device_infos by port
  """
  map_usb_hub_ports = {}
  for device_info in device_info_list:
    hub_address = device_info['usb_hub_address']
    port = device_info['usb_hub_port']
    if hub_address:
      if hub_address not in map_usb_hub_ports:
        map_usb_hub_ports[hub_address] = {}
      if not map_usb_hub_ports[hub_address].get(
          port) or device_info['ftdi_interface'] == 2:
        map_usb_hub_ports[hub_address][port] = device_info
  return map_usb_hub_ports


def _discover_cambrionix_devices(manager, cambrionix_info_list,
                                 device_info_list):
  """Discover all of the devices that are connected to a cambrionix.

  For each Cambrionix that is in the cambrionix_info_list, this function will
  use the devices in the device list to create port_list with information about
  each device that is connected to the Cambrionix USB hub.

  Args:
      manager (Manager): instance of Manager class
      cambrionix_info_list (list): list of devices for Cambrionix USB hubs
      device_info_list (list): list of devices for devices that have been
        detected

  Note: See the _identify_device function for more information on the
    device_data dicts that are used for the cambrionix_info_list and
    device_info_list.
  """
  device_config = {}
  map_usb_hub_ports = _get_usb_hub_map(device_info_list)
  for cambrionix_info in cambrionix_info_list:
    cambrionix_info['is_supported_hub'] = False
    try:
      device_config = {
          'persistent': {
              'console_port_name': cambrionix_info['address'],
              'name': '{}'.format(cambrionix_info['name']),
              'serial_number': cambrionix_info['serial_number']
          },
          'options': {},
          'make_device_ready': 'on',
          'log_name_prefix': 'port-map'
      }
      # Create the hub object. We don't use manager here since this is used
      # for hubs that have not been detected.
      cambrionix_inst = cambrionix.Cambrionix(
          manager=manager,
          device_config=device_config,
          log_file_name=None,
          log_directory=manager.log_directory  # pytype: disable=attribute-error
      )
      cambrionix_inst.get_detection_info()  # populate undetected ones.
    except (KeyError, errors.DeviceError):
      cambrionix_info['port_list'] = []
      cambrionix_info['port_count'] = 0
      continue

    try:
      cambrionix_info['port_count'] = cambrionix_inst.total_ports
      cambrionix_info['port_list'] = [None] * cambrionix_info['port_count']
      port_map = map_usb_hub_ports.get(cambrionix_info['address'], {})
      for port, device_info in port_map.items():
        if not port:
          continue
        cambrionix_info['port_list'][port-1] = device_info
      cambrionix_info['is_supported_hub'] = True
      cambrionix_info[
          'port_modes'] = cambrionix_inst.switch_power.get_all_ports_mode()
    finally:
      cambrionix_inst.close()


class UsbPortMap(object):
  """Class for discovering USB information for connected devices, and printing a port-map."""

  def __init__(self, manager):
    """Initialize the USB port map instance by discovering devices and USB hubs.

    Args:
        manager (Manager):  instance of Manager class
    """
    self.manager_weakref = weakref.ref(manager)
    self.cambrionix_info_list = []
    self.device_info_list = []
    for usb_info_inst in usb_utils.get_address_to_usb_info_dict().values():
      device_info = _identify_device(usb_info_inst)
      if device_info:
        if 'cambrionix' in device_info.get('name', ''):
          self.cambrionix_info_list.append(device_info)
        else:
          self.device_info_list.append(device_info)

    _discover_cambrionix_devices(self.manager_weakref(),
                                 self.cambrionix_info_list,
                                 self.device_info_list)

    # remove devices associated with a cambrionix from the device list
    self.device_info_list = [
        device_info for device_info in self.device_info_list
        if not device_info['usb_hub_address']
    ]

    # use what has been detected to add to what was discovered
    # pytype: disable=attribute-error
    gdm_config = self.manager_weakref()._devices
    # pytype: enable=attribute-error
    self._correlate_discovered_and_detected(gdm_config)
    self._get_detected_device_information(gdm_config)
    self._get_power_state_information()
    self._check_for_conflicts(gdm_config)

  def get_port_map(self):
    """Get a dictionary containing the port map information for the discovered devices.

    Returns:
        list: A list of dictionaries. Each entry in the list will be a hub
        or a device. A hub will have a list of devices by port number, and a
        device will have the port-map fields for that device, similar to what is
        printed by print_port_map.
    """
    port_map = []
    for cambrionix_info in self.cambrionix_info_list:
      port_map.append(self._get_hub_and_device_information(cambrionix_info))

    for device_info in self.device_info_list:
      other_info = self._get_other_device_information(device_info)
      if other_info:
        port_map.append(other_info)

    return port_map

  def print_port_map(self):
    """Print the port map based on what was discovered when the UsbPortMap object was created."""
    port_map = self.get_port_map()

    # Print the information for each hub in the port map
    for hub_or_device in port_map:
      if hub_or_device['hub_device']:
        hub_header_printed = False
        if 'port_list' not in hub_or_device:
          continue
        for port_index, single in enumerate(hub_or_device['port_list']):
          if not single:
            continue
          if not hub_header_printed:
            hub_header_printed = True
            if hub_or_device['is_detected']:
              logger.info('\n{}'.format(hub_or_device['hub_name']))
            else:
              logger.info('\n{} - Undetected'.format(hub_or_device['hub_name']))
            logger.info('-' * len(hub_or_device['hub_name']))
            logger.info('\n{}\n'.format(HUB_DEVICES_COLUMN_HEADER))

          logger.info('{:>2}: {:13} {:7} {:25} {}'.format(
              port_index+1, single['Issue'], single['Mode'],
              single['Discovered'], single['Configured']))

    # Now that the hubs have been printed, loop through and print single devices
    other_header_printed = False
    for hub_or_device in port_map:
      if not hub_or_device['hub_device']:
        if not other_header_printed:
          other_header_printed = True
          logger.info('\n{}'.format(OTHER_DEVICES_HEADER))
          logger.info('-' * len(OTHER_DEVICES_HEADER))
          logger.info('\n{}\n'.format(OTHER_DEVICES_COLUMN_HEADER))

        logger.info('    {:13} Sync    {}'.format(hub_or_device['Issue'],
                                                  hub_or_device['Discovered']))

  def _get_hub_and_device_information(self, hub):
    """Get information for all devices connected to a known USB hub.

    Args:
        hub (dict): A dictionary containing the information for the USB hub

    Returns:
        dict: A dictionary containing just the port map information for the
        hub, similar to what is printed for print_port_map.
    """
    hub_dict = {}
    hub_dict['hub_name'] = hub['name']
    hub_dict['hub_device'] = True
    hub_dict['is_detected'] = hub['is_detected']
    if not hub['is_supported_hub']:
      hub_dict['port_count'] = 0
      hub_dict['port_list'] = []
      return hub_dict

    hub_dict['port_list'] = [None] * (hub['port_count'] + 1)
    hub_dict['port_count'] = hub['port_count']

    for port, device in enumerate(hub['port_list']):
      if device:
        # Based on the information on data in device, determine what will be printed
        mode_string = 'N/A'
        if 'port_mode' in device:
          mode_string = device['port_mode']
        device_string = device['description']
        if 'gdm_config_name' in device:
          device_string = device['gdm_config_name']
        if device['type']:
          device_string += ' ' + device['type']
        configured_devices = ','.join(device['detected_devices'])
        issue_string = ''
        if configured_devices:
          if len(device['detected_devices']) > 1:
            issue_string = 'Conflict'
          if 'gdm_config_name' in device and \
                  device['gdm_config_name'] not in configured_devices:
            issue_string = 'Conflict'
        elif 'gdm_config_name' in device:
          issue_string = 'NotConfigured'
        if not issue_string and NOT_DETECTED_CONFLICT in device[
            'conflict_types']:
          issue_string = 'Undetected'

        hub_dict['port_list'][port] = {
            'Issue': issue_string,
            'Mode': mode_string,
            'Discovered': device_string,
            'Configured': configured_devices,
        }
      else:
        logger.info('{:>2}: -'.format(port))

    return hub_dict

  def _get_other_device_information(self, device):
    """Get the device information for a single device that is not associated with a USB hub.

    Args:
        device (dict): A dictionary containing information for a single device.

    Returns:
        dict: A dictionary containing device information for port-map or None.
    """
    if not device['known_device']:
      return None

    device_string = device['description']
    if 'gdm_config_name' in device:
      device_string = device['gdm_config_name']
    if device['type']:
      device_string += ' ' + device['type']
    issue_string = ''
    # pytype: disable=key-error
    if USB_PORT_CONFLICT in device['conflict_types']:
      issue_string = 'Conflict'
    if USB_HUB_CONFLICT in device['conflict_types']:
      issue_string = 'Conflict'
    if NOT_DETECTED_CONFLICT in device['conflict_types']:
      issue_string = 'Undetected'
    # pytype: enable=key-error

    device_dict = {}
    device_dict['hub_device'] = False
    device_dict['Issue'] = issue_string
    device_dict['Mode'] = 'Sync'
    device_dict['Discovered'] = device_string
    device_dict['Configured'] = ''
    return device_dict

  def _check_known_devices(self, device, serial_numbers, addresses):
    """Checks if usb connection matches a known device.

    Args:
        device (dict): A dictionary containing information about an
          individual device.
        serial_numbers (dict): The devices from GDM config with a configured
          serial number.
        addresses (dict): The devices from GDM config with a configured
          address.
    """
    if device:
      if device['address'] in addresses:
        device['gdm_config_name'] = addresses[device['address']]
      elif device['serial_number'] in serial_numbers:
        device['gdm_config_name'] = serial_numbers[device['serial_number']]
        del serial_numbers[device['serial_number']]

  def _correlate_discovered_and_detected(self, gdm_config):
    """Use the GDM config for detected devices to identify the discovered devices.

    Args:
        gdm_config (dict): a dictionary containing the devices detected by GDM.
    """
    serial_numbers = {}
    addresses = {}
    for name in gdm_config:
      if 'serial_number' in gdm_config[name]['persistent']:
        serial_numbers[gdm_config[name]['persistent']['serial_number']] = name
      if 'console_port_name' in gdm_config[name]['persistent']:
        addresses[gdm_config[name]['persistent']['console_port_name']] = name
    # pytype: disable=attribute-error
    other_devices = self.manager_weakref().other_devices
    # pytype: enable=attribute-error

    for cambrionix_info in self.cambrionix_info_list:
      cambrionix_info['is_detected'] = False
      if cambrionix_info['name'] in other_devices:
        cambrionix_info['is_detected'] = True

      for device_info in cambrionix_info['port_list']:
        self._check_known_devices(device_info, serial_numbers, addresses)

    for device_info in self.device_info_list:
      self._check_known_devices(device_info, serial_numbers, addresses)

    missing_device_info_list = []
    for number in serial_numbers:
      if serial_numbers[number] not in missing_device_info_list:
        missing_device_info_list.append(serial_numbers[number])

    for number in addresses:
      if addresses[number] not in missing_device_info_list:
        missing_device_info_list.append(addresses[number])

    # Some devices have a single connection, but may have an entry in more than
    # one list. For example, some devices have a serial number and an ADB serial
    # number. Iterate over the discovered devices and remove the extra entry so
    # that they are not reported as a missing device.
    for cambrionix_info in self.cambrionix_info_list:
      for device in cambrionix_info['port_list']:
        if not device:
          continue

        if ('gdm_config_name' in device and
            gdm_config[device['gdm_config_name']]['persistent']['device_type']
            in SINGLE_CONNECTION_LIST):
          if device['gdm_config_name'] in missing_device_info_list:
            missing_device_info_list.remove(device['gdm_config_name'])

    for device in self.device_info_list:
      if ('gdm_config_name' in device and gdm_config[device['gdm_config_name']]
          ['persistent']['device_type'] in SINGLE_CONNECTION_LIST):
        missing_device_info_list.remove(device['gdm_config_name'])

    self.missing_devices = missing_device_info_list

  def _check_for_conflicts(self, gdm_config):
    """Use the GDM config for detected devices to add information or look for conflicts.

    Args:
        gdm_config (dict): a dictionary containing the devices detected by GDM
    """
    # Look for conflicts for devices that were discovered and are connected to a cambrionix
    for cambrionix_info in self.cambrionix_info_list:
      for index, device in enumerate(cambrionix_info['port_list']):
        if not device:
          continue

        if device['known_device'] and 'gdm_config_name' not in device:
          device['conflict_types'].append(NOT_DETECTED_CONFLICT)

        if 'gdm_config_name' in device:
          device_id = device['gdm_config_name']
          usb_hub_name, usb_port = self._get_usb_hub_info(
              device['type'], device_id, gdm_config)
          if usb_hub_name and usb_hub_name != cambrionix_info['name']:
            device['configured_hub'] = usb_hub_name
            device['conflict_types'].append(USB_HUB_CONFLICT)

          if usb_port and usb_port != str(index):
            device['configured_port'] = usb_port
            device['conflict_types'].append(USB_PORT_CONFLICT)

          if device['type'] == 'USB':
            if not usb_hub_name:
              device['conflict_types'].append(MISSING_USB_HUB)
            if not usb_port:
              device['conflict_types'].append(MISSING_USB_PORT)

    # Look for conflicts for devices that were discovered and are not connected
    # to a Cambrionix.
    for device in self.device_info_list:
      if device['known_device'] and 'gdm_config_name' not in device:
        device['conflict_types'].append(  # pytype: disable=key-error
            NOT_DETECTED_CONFLICT)
      if 'gdm_config_name' in device:
        device_id = device['gdm_config_name']
        usb_hub_name, usb_port = self._get_usb_hub_info(device['type'],
                                                        device_id, gdm_config)
        if usb_hub_name:
          device['conflict_types'].append(USB_HUB_CONFLICT)
        if usb_port:
          device['conflict_types'].append(USB_PORT_CONFLICT)

  def _get_usb_hub_info(self, device_type, device_id, gdm_config):
    """Get the configured USB hub name and port for the specified device.

    Args:
        device_type (str): type of the device.
        device_id (str): the GDM name for the device.
        gdm_config (dict): a dictionary containing the devices detected by GDM.

    Returns:
        tuple: (USB hub name, USB port number).
        (None, None) if they are not set.
    """
    return (gdm_config[device_id]['persistent'].get('device_usb_hub_name')
            or gdm_config[device_id]['options'].get('usb_hub'),
            gdm_config[device_id]['persistent'].get('device_usb_port')
            or gdm_config[device_id]['options'].get('usb_port'))

  def _get_detected_device_information(self, gdm_config):
    """Use the GDM config for detected devices to add information to the device dictionary.

    Args:
        gdm_config (dict): a dictionary containing the devices detected by GDM
    """
    cambrionix_lookup = {}
    for index, cambrionix_info in enumerate(self.cambrionix_info_list):
      # usb_hub_name could be the cambrionix device name or serial port path.
      cambrionix_lookup[cambrionix_info['address']] = index
      cambrionix_lookup[cambrionix_info['name']] = index

    # Add the detected device to the device at cambrionix/port if there is a
    # mismatch
    for gdm_device in gdm_config:
      device_types = (
          'usb',
      )
      usb_hub_infos = (
          self._get_usb_hub_info(device_type, gdm_device, gdm_config)
          for device_type in device_types)
      for usb_hub_name, usb_port in usb_hub_infos:
        if usb_hub_name and usb_port and usb_hub_name in cambrionix_lookup:
          hub_index = cambrionix_lookup[usb_hub_name]
          index = int(usb_port) - 1
          if index < len(self.cambrionix_info_list[hub_index]['port_list']):
            if not self.cambrionix_info_list[hub_index]['port_list'][index]:
              self.cambrionix_info_list[hub_index]['port_list'][index] = {}
              device = self.cambrionix_info_list[hub_index]['port_list'][index]
              _initialize_default_device(device)
              device['conflict_types'].append(NOT_DISCOVERED_CONFLICT)
              device['detected_devices'].append(gdm_device)
            else:
              device = self.cambrionix_info_list[hub_index]['port_list'][index]
              device['detected_devices'].append(gdm_device)
              if device.get('gdm_config_name') != gdm_device:
                device['conflict_types'].append(NOT_DISCOVERED_CONFLICT)

  def _get_power_state_information(self):
    """For devices connected to a Cambrionix USB hub, get the power state mode.

    Note:
        This method should be called after _discover_cambrionix_devices.
        This method uses the the port state in the cambrionix device that is set
        by _discover_cambrionix_devices.
    """
    for cambrionix_info in self.cambrionix_info_list:
      if not cambrionix_info['is_supported_hub']:
        continue
      for index, device in enumerate(cambrionix_info['port_list']):
        if not device:
          cambrionix_info['port_list'][index] = {}
          device = cambrionix_info['port_list'][index]
          _initialize_default_device(device)

        device['port_mode'] = cambrionix_info['port_modes'][index]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for GDM logger."""
import atexit
import logging
import logging.handlers
import multiprocessing as mp
import os
import re
import sys
import types

from gazoo_device import config
from gazoo_device import multiprocess_logging

# Making this global enables user control of stdout streaming (indirectly)
_stdout_handler = None

# Log formats for debug
FMT = ('%(asctime)s.%(msecs)03d %(levelname).1s %(process)5d '
       '%(filename)19.19s:%(lineno)d\t%(message)s')
DATEFMT = '%Y%m%d %X'

dirname = os.path.dirname(config.DEFAULT_LOG_FILE)
if not os.path.isdir(dirname):
  os.makedirs(dirname)


def add_handler(handler):
  """Adds a logging handler to the LoggingThread.

  It receives messages sent to GDM Loggers in the main process and its child
  processes.

  Along with remove_handler, allows library users to configure extra
  destinations and formats for log messages emitted from within GDM.

  Args:
      handler (logging.handler): A logging handler
  """
  logger = get_logger()
  logger.logging_thread.add_handler(handler)  # pytype: disable=attribute-error


def create_queue_handler(log_level):
  """Adds a QueueHandler to the top-level 'gazoo_device_manager' logger.

  Should be used once per process, and will be the only handler on a GDM
  logger object in a given process. Destination handlers are added instead to
  the LoggingThread object via add_handler().

  Args:
      log_level (int): Integer log level, e.g. logging.DEBUG (value is 10)
  """
  logger = get_logger()
  handler = multiprocess_logging.QueueHandler(
      logger.logging_queue)  # pytype: disable=attribute-error
  handler.setLevel(log_level)
  logger.handlers = [handler]


def flush_queue_messages():
  """Wait until all messages currently in the logger queue are flushed."""
  get_logger().logging_thread.sync()  # pytype: disable=attribute-error


def get_logger(component_name=None):
  """Returns a Logger that inherits from (or is) the top-level GDM Logger.

  Differs from usual getLogger in that the name given is appended to the name
  of the top-level GDM Logger (e.g. get_logger('component') is equivalent
  to logging.getLogger('gazoo_device_manager.component')), and also in that
  message strings passed to the returned Logger will be combined with the
  given args using .format() rather than %.

  Args:
      component_name (str): name of an GDM component. Nests using '.' char.

  Returns:
      logging.Logger: main logger or sub logger.
  """
  if component_name is not None:
    name = '.'.join(['gazoo_device_manager', component_name])
  else:
    name = 'gazoo_device_manager'

  logger = logging.getLogger(name)

  # Override _log with our own _brace_format_log
  if not hasattr(logger, '_original_log'):
    logger._original_log = logger._log
    logger._log = types.MethodType(_brace_format_log, logger)

  return logger


def initialize_logger():
  """Configures the top-level GDM Logger.

  Configures it to begin logging to stdout and to the default log destination
  (config.DEFAULT_LOG_FILE).
  """
  # Set up multiprocess logging thread
  logging_queue = mp.Queue(-1)
  logging_queue.cancel_join_thread()
  logging_thread = multiprocess_logging.LoggingThread(logging_queue)
  atexit.register(logging_thread.stop)

  # Set up core GDM logger
  logger = get_logger()
  logger.setLevel(logging.DEBUG)
  logger.logging_queue = logging_queue
  logger.logging_thread = logging_thread

  # Create handler for core GDM logger in main process
  create_queue_handler(logging.DEBUG)

  # Configure a handler that writes to GDM logfile
  filepath = config.DEFAULT_LOG_FILE
  args = dict(mode='a', maxBytes=100 * 1024 * 1023, backupCount=5)
  logfile_handler = logging.handlers.RotatingFileHandler(filepath, **args)
  logfile_handler.setLevel(logging.DEBUG)
  logfile_formatter = logging.Formatter(FMT, datefmt=DATEFMT)
  logfile_handler.setFormatter(logfile_formatter)
  add_handler(logfile_handler)

  # Configure a handler that writes INFO logs to stdout
  stdout_handler = logging.StreamHandler(sys.stdout)
  stdout_handler.setLevel(logging.INFO)
  stdout_formatter = logging.Formatter('%(message)s')
  stdout_handler.setFormatter(stdout_formatter)
  add_handler(stdout_handler)

  # Keep global copy of stdout_handler created and added above to allow for
  # changing the log level later for that handler.
  global _stdout_handler
  _stdout_handler = stdout_handler

  logging_thread.start()


def remove_handler(handler):
  """Removes the given handler from the GDM Logger's LoggingThread.

  Messages sent to the GDM Logger will no longer be sent to that handler.

  Along with add_handler, allows library users to configure extra destinations
  and formats for log messages emitted from within GDM.

  Args:
      handler (logging.handler): A logging handler that was added earlier
        using add_handler
  """
  logger = get_logger()
  logger.logging_thread.remove_handler(handler)


def set_component_log_level(component_name, log_level):
  """Sets the log level for the named component.

  Args:
      component_name (str): name of an GDM component
      log_level (int): Integer log level, e.g. logging.DEBUG (value is 10)
  """
  logger = get_logger(component_name)
  logger.setLevel(log_level)


def silence_progress_messages():
  """Stops the default behavior of streaming GDM Logger messages to stdout."""
  if _stdout_handler:
    remove_handler(_stdout_handler)


def stream_debug():
  """Sets the log level for GDM Logger stdout streaming to DEBUG."""
  if _stdout_handler:
    fmt = logging.Formatter(FMT, datefmt=DATEFMT)
    _stdout_handler.setFormatter(fmt)
    _stdout_handler.setLevel(logging.DEBUG)


def _brace_format_log(self,
                      level,
                      msg,
                      args,
                      exc_info=None,
                      extra=None,
                      **kwargs):
  """Enables brace logging for GDM Loggers (by overriding Logger._log)."""
  # Combines msg with args and kwargs using .format(),
  # saving special kwargs exc_info and extra for original _log method
  if args or kwargs:
    formatted_msg = str(msg).format(*args, **kwargs)
  else:
    formatted_msg = str(msg)

  new_kwargs = dict(exc_info=exc_info, extra=extra)

  self._original_log(level, formatted_msg, (), **new_kwargs)


class LogData(object):
  """Stores data to be added to log lines and automatically formatted."""

  key_pattern = re.compile(r'\W')

  def __init__(self, **items):
    self.items = items

  def __str__(self):
    r"""Formats data items into the form '##\tkey1=value1\tkey2=value2'.

    In keys, strips non-alphanumeric characters.
    In values, replaces escape char / and delimiters \t, = with /0, /1, /2.

    Returns:
      str: data as string.
    """
    strings = ['##']

    for k, v in self.items.items():
      key = self.key_pattern.sub('', str(k))
      val = str(v).replace('/', '/0').replace('\t', '/1').replace('=', '/2')
      key_val_string = '='.join([key, val])
      strings.append(key_val_string)

    return '\t'.join(strings)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for validating architecture conformance."""
import inspect
import re
from typing import Any, Callable, Collection, List, NoReturn, Optional, Tuple
from typing import Type, Union
from gazoo_device import decorators
from gazoo_device import manager
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.base_classes import auxiliary_device_base
from gazoo_device.base_classes import gazoo_device_base
from gazoo_device.base_classes import primary_device_base
from gazoo_device.capabilities.interfaces import capability_base

_ALLOWED_PROPERTY_TYPES_CAPABILITY = (decorators.PersistentProperty,
                                      decorators.DynamicProperty,
                                      decorators.OptionalProperty)
_ALLOWED_PROPERTY_TYPES_DEVICE = (decorators.PersistentProperty,
                                  decorators.DynamicProperty,
                                  decorators.OptionalProperty,
                                  decorators.CapabilityProperty)
_LOG_DECORATOR_BASE_CLASSES = (auxiliary_device.AuxiliaryDevice,
                               gazoo_device_base.GazooDeviceBase,
                               capability_base.CapabilityBase)
_PUBLIC_METHOD_BASE_CLASSES = (primary_device_base.PrimaryDeviceBase,
                               gazoo_device_base.GazooDeviceBase,
                               auxiliary_device.AuxiliaryDevice,
                               auxiliary_device_base.AuxiliaryDeviceBase)
_REQUIRED_DEVICE_CLASS_CONSTANTS = ("DEVICE_TYPE",
                                    "COMMUNICATION_TYPE",
                                    "DETECT_MATCH_CRITERIA",
                                    "_OWNER_EMAIL")
_RETURN_SECTION_MULTILINE = r"\n\s*Returns:\s*\n"
_RETURN_SECTION_ONELINE = r"\b(R|r)eturn(s)?\b"
_MISMATCHING_SIGNATURE_TEMPLATE = """\tMethod: {}
\t\tInherited signature:
\t\t{}
\t\tConflicting signature(s) inherited from:
\t\t{}."""

_LogDecoratorType = Union[Type[auxiliary_device.AuxiliaryDevice],
                          Type[gazoo_device_base.GazooDeviceBase],
                          Type[capability_base.CapabilityBase]]


def get_invalid_capability_definitions(
    device_class: Type[Any]) -> List[Tuple[str, str]]:
  """Returns all invalid capability definitions in the device class."""
  invalid_definitions = []
  for attr_name, attr in inspect.getmembers(device_class):
    if isinstance(attr, decorators.CapabilityProperty):
      try:
        _validate_capability_definition(attr)
      except (TypeError, ValueError, RuntimeError) as e:
        invalid_definitions.append((attr_name, repr(e)))
  return invalid_definitions


def get_invalid_health_check_names(device_class: Type[Any]) -> List[str]:
  """Returns names of health checks which do not start with "check_<...>"."""
  incompliant_attrs = []
  for attr_name, attr in inspect.getmembers(device_class):
    if (callable(attr)
        and hasattr(attr, "__health_check__") and
        not attr_name.startswith("check_")):
      incompliant_attrs.append(attr_name)
  return incompliant_attrs


def get_invalid_public_methods(
    device_class: Type[Any],
    excluded_methods: Optional[Collection[str]] = None) -> List[str]:
  """Returns names of public methods which are not allowed.

  The only allowed public methods are:
    * the ones required by base_classes/primary_device_base.py;
    * the ones added in base_classes/gazoo_device_base.py;
    * health checks;
    * deprecated aliases.
  All other public functionality in device classes must be implemented as
  capabilities.

  Args:
    device_class: Device class to check.
    excluded_methods: Names of methods to exclude from this check.

  Returns:
    Names of incompliant attribute names.
  """
  excluded_methods = excluded_methods or []
  incompliant_attrs = []
  base_attrs = set()
  for base_class in _PUBLIC_METHOD_BASE_CLASSES:
    base_attrs |= set(vars(base_class))

  for attr_name, attr in inspect.getmembers(device_class):
    if (not attr_name.startswith("_")
        and callable(attr)
        and not hasattr(attr, "__deprecated__")
        and not hasattr(attr, "__health_check__")
        and attr_name not in base_attrs
        and attr_name not in excluded_methods):
      incompliant_attrs.append(attr_name)
  return incompliant_attrs


def get_log_decorator_violators(cls: _LogDecoratorType,
                                check_inherited: bool = True) -> List[str]:
  """Returns method names which do not comply with the log decorator.

  Every public method without a return value must be decorated with
  @decorators.LogDecorator(<logger>).

  Args:
    cls: Class inheriting from AuxiliaryDevice, GazooDeviceBase, or
      CapabilityBase.
    check_inherited: Whether inherited methods need to be checked for
      compliance.

  Returns:
    Names of incompliant methods.

  Raises:
    TypeError: Invalid class provided.
  """
  if not issubclass(cls, _LOG_DECORATOR_BASE_CLASSES):
    raise TypeError(
        "Can only check log decorator API compliance for instances of "
        "{}. Found type: {}.".format(_LOG_DECORATOR_BASE_CLASSES, cls))

  incompliant_methods = []
  if check_inherited:
    attribute_names = dir(cls)
  else:
    attribute_names = vars(cls).keys()

  for attribute_name in attribute_names:
    attribute = getattr(cls, attribute_name)
    if _is_decoration_required(attribute):
      incompliant_methods.append(attribute_name)
  return incompliant_methods


def get_mismatching_signatures(
    cls: Type[Any],
    excluded_methods: Optional[Collection[str]] = None
) -> List[Tuple[str, str, List[str]]]:
  """Returns methods that do not include same args as the parent method.

  In other words, if a parent defines a method 'some_method' and the given
  class (cls) also defines the same method, add 'some_method' to list of
  mismatching signatures if cls.some_method signature does not start with the
  parent.some_method signature. Ignores __init__ and __new__ signatures.

  Args:
    cls: Class object to check methods signatures of.
    excluded_methods: Method names for which to ignore mismatched signatures.

  Returns:
    List of tuples, where each entry is
    (method name,
     class with method definition and method signature,
     [list of parents with different method signatures and signatures]).
  """
  excluded_methods = excluded_methods or []
  parents = cls.__mro__[1:]
  mismatches = []

  for attr_name, attr in inspect.getmembers(cls):
    if (inspect.isfunction(attr) or inspect.ismethod(attr)):
      if attr_name in ["__init__", "__new__"]:
        continue
      child_method = decorators.unwrap(attr)
      child_signature = inspect.getfullargspec(child_method).args

      parents_with_method = [
          parent for parent in parents if attr_name in parent.__dict__
      ]
      if attr_name in cls.__dict__:
        signature_defined_in = cls
      else:
        signature_defined_in = parents_with_method[0]

      parents_with_conflicting_signature = []
      for parent in parents_with_method:
        parent_method = decorators.unwrap(getattr(parent, attr_name))
        parent_signature = inspect.getfullargspec(parent_method).args

        len_parent = len(parent_signature)
        if child_signature[:len_parent] != parent_signature:
          parents_with_conflicting_signature.append("{}.{}.{}({})".format(
              parent.__module__, parent.__name__, attr_name,
              ", ".join(parent_signature)))

      if (parents_with_conflicting_signature
          and attr_name not in excluded_methods):
        mismatches.append(
            (attr_name,
             "{}.{}.{}({})".format(signature_defined_in.__module__,
                                   signature_defined_in.__name__,
                                   attr_name,
                                   ", ".join(child_signature)),
             parents_with_conflicting_signature))
  return mismatches


def get_mismatching_signature_err_strs(
    classes: Collection[Type[Any]],
    excluded_methods: Optional[Collection[str]] = None) -> List[str]:
  """Returns a list of the errors for the mismatched signatures on the classes.

  Looks for class signatures that don't match their parent's signatures.
  Unlike normal linting, it allows for children to add additional args as long
  as they start with the parent's args.

  Args:
    classes: Classes to look for mismatching signatures in their methods.
    excluded_methods: Method names for which to ignore mismatched signatures.

  Returns:
    Error messages identifying classes and their methods with mismatched
    signatures.
  """
  mismatches_strs = []
  for cls in classes:
    mismatching_signatures = get_mismatching_signatures(cls, excluded_methods)
    if mismatching_signatures:
      mismatching_signatures_str = "\n".join(
          _MISMATCHING_SIGNATURE_TEMPLATE.format(
              method, child_sig, "\n\t\t".join(parent_sigs))
          for method, child_sig, parent_sigs in mismatching_signatures)
      mismatches_strs.append("Class {}:\n{}".format(cls,
                                                    mismatching_signatures_str))
  return mismatches_strs


def get_uncategorized_properties(
    cls: Type[Any],
    excluded_properties: Optional[Collection[str]] = None) -> List[str]:
  """Returns names of public properties which do not fall into a known category.

  Allowed categories: persistent, dynamic, optional properties.
  Device classes can also define capabilities.

  Args:
    cls: Class object to check.
    excluded_properties: Names of properties to exclude from this check.

  Returns:
    Names of incompliant properties.
  """
  excluded_properties = excluded_properties or []
  if issubclass(cls, capability_base.CapabilityBase):
    allowed_property_types = _ALLOWED_PROPERTY_TYPES_CAPABILITY
  else:
    allowed_property_types = _ALLOWED_PROPERTY_TYPES_DEVICE
  incompliant_properties = []
  for attr_name, attr in inspect.getmembers(cls):
    if (not attr_name.startswith("_")
        and isinstance(attr, property)
        and attr_name not in excluded_properties
        and not hasattr(attr.fget, "__deprecated__")
        and not isinstance(attr, allowed_property_types)):
      incompliant_properties.append(attr_name)
  return incompliant_properties


def get_unset_constants(device_class: Type[Any]) -> List[str]:
  """Returns names of unset required class constants."""
  return [attr for attr in _REQUIRED_DEVICE_CLASS_CONSTANTS
          if not getattr(device_class, attr, None)]


def _validate_capability_definition(
    capability_property: decorators.CapabilityProperty) -> None:
  """Raises an error if the capability definition is invalid."""
  flavor_classes = list(capability_property.capability_classes)
  all_flavors = manager.Manager.get_all_supported_capability_flavors().values()
  for flavor_class in flavor_classes:
    if not inspect.isclass(flavor_class):
      raise TypeError("Object {} is not a class object.".format(flavor_class))
    if not issubclass(flavor_class, capability_base.CapabilityBase):
      raise TypeError(
          "Object {} is not a subclass of CapabilityBase.".format(flavor_class))
    if flavor_class not in all_flavors:
      raise ValueError(
          "Capability class {} is not supported. "
          "Supported capability classes: {}.".format(flavor_class, all_flavors))

  expected_capability_interface = flavor_classes[0].get_capability_interface()
  for flavor_class in flavor_classes[1:]:
    capability_interface = flavor_class.get_capability_interface()
    if capability_interface != expected_capability_interface:
      raise RuntimeError(
          "Not all capability flavors share the same capability interface. "
          "Differing interfaces: {}, {}."
          .format(expected_capability_interface, capability_interface))

  expected_name = expected_capability_interface.get_capability_name()
  if capability_property.fget.__name__ != expected_name:
    raise RuntimeError(
        "Attempting to define capability flavor(s) {} under invalid name {}. "
        "Capability interface: {}, expected name: {}.".format(
            flavor_classes, capability_property.fget.__name__,
            expected_capability_interface, expected_name))


def _is_instance_method(obj: Any) -> bool:
  """Returns True if obj is an instance method (bound or unbound)."""
  # "Unbound" methods are considered functions in Python 3
  is_method_or_func = inspect.ismethod(obj) or inspect.isfunction(obj)

  if is_method_or_func:
    unwrapped_func = decorators.unwrap(obj)
    args = inspect.getfullargspec(unwrapped_func).args
    return bool(args) and args[0] == "self"

  return False


def _is_decoration_required(attr: Any) -> bool:
  """Check if log decorator is required for this attribute.

  Methods that need to be decorated are public class instance methods which do
  not return a value.

  Args:
    attr: Attribute to check.

  Returns:
    Whether the attribute should be decorated with the log decorator.
  """
  return (_is_instance_method(attr)  # Bound or unbound instance method
          and not attr.__name__.startswith("_")  # Public
          and not getattr(attr, "__isabstractmethod__", False)  # Not abstract
          and not _has_return_value(attr)  # Has a return other than None
          and decorators.WRAPS_ATTR_NAME not in attr.__dict__)  # Not decorated


def _has_return_value(method: Callable[..., Any]) -> Optional[bool]:
  """Returns True if the method returns anything other than None.

  Args:
    method: Method to check.

  Returns:
    bool: True if the method has a return value, False otherwise.
    None: If it's not possible to tell (no docstring and no type annotation).
  """
  has_return_value = _has_return_value_from_type_annotation(method)
  if has_return_value is None:  # No return type annotation
    has_return_value = _has_return_value_from_docstring(method)
  return has_return_value


def _has_return_value_from_docstring(
    method: Callable[..., Any]) -> Optional[bool]:
  """Parses the docstring to determine if the method returns a value.

  Args:
    method: Method to check.

  Returns:
    bool: True if the method has a return value, False otherwise.
    None: If it's not possible to tell (no docstring).
  """
  docstring = method.__doc__
  if docstring is None:
    return None
  if len(docstring.splitlines()) > 1:  # Multi-line docstring
    return re.search(_RETURN_SECTION_MULTILINE, docstring) is not None
  else:
    return re.search(_RETURN_SECTION_ONELINE, docstring) is not None


def _has_return_value_from_type_annotation(
    method: Callable[..., Any]) -> Optional[bool]:
  """Examines the type annotation to determine if method returns.

  Args:
    method: method to check.

  Returns:
    bool: True if the method has a return value, False otherwise.
    None: If it's not possible to tell (no return value type annotation).
  """
  type_annotations = inspect.getfullargspec(method).annotations
  if "return" not in type_annotations:
    return None
  return type_annotations["return"] not in [None, NoReturn]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Manager module.

  - detects devices
  - creates devices
  - get props and sets optional props
"""
import atexit
import copy
import datetime
import difflib
import fnmatch
import inspect
import json
import logging
import multiprocessing
import os
import queue
import re
import shutil
import signal
import subprocess
import textwrap
import time
from typing import Dict, Optional, Union

from gazoo_device import config
from gazoo_device import custom_types
from gazoo_device import device_detector
from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import gdm_logger

from gazoo_device.capabilities import event_parser_default
from gazoo_device.log_parser import LogParser
from gazoo_device.switchboard import communication_types
from gazoo_device.switchboard import switchboard

from gazoo_device.usb_port_map import UsbPortMap
from gazoo_device.utility import common_utils
from gazoo_device.utility import host_utils
from gazoo_device.utility import parallel_utils
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_logger()


class Manager():
  """Manages the setup and communication of smart devices."""

  def __init__(self,
               device_file_name=None,
               device_options_file_name=None,
               testbeds_file_name=None,
               gdm_config_file_name=config.DEFAULT_GDM_CONFIG_FILE,
               log_directory=None,
               gdm_log_file=None,
               gdm_log_formatter=None,
               adb_path=None,
               debug_level=logging.DEBUG,
               stream_debug=False,
               stdout_logging=True,
               max_log_size=100000000):

    self._open_devices = {}
    self.max_log_size = max_log_size
    # b/141476623: exception queue must not share multiprocessing.Manager()
    common_utils.run_before_fork()
    self._exception_queue_manager = multiprocessing.Manager()
    common_utils.run_after_fork_in_parent()
    self._exception_queue = self._exception_queue_manager.Queue()

    # Backwards compatibility for older debug_level=string style __init__
    if not isinstance(debug_level, int):
      if debug_level in ["debug", "developer"]:
        debug_level = logging.DEBUG
      else:
        debug_level = logging.INFO
    logger.level = debug_level

    if stream_debug:
      gdm_logger.stream_debug()

    if not stdout_logging:
      gdm_logger.silence_progress_messages()

    if gdm_log_file:
      if not gdm_log_formatter:
        gdm_log_formatter = logging.Formatter(
            gdm_logger.FMT, datefmt=gdm_logger.DATEFMT)

      self.gdm_log_handler = logging.FileHandler(gdm_log_file)
      self.gdm_log_handler.setLevel(debug_level)
      self.gdm_log_handler.setFormatter(gdm_log_formatter)
      gdm_logger.add_handler(self.gdm_log_handler)

    self.device_file_name = None
    self.device_options_file_name = None
    self.testbeds_file_name = None
    self.log_directory = None
    self._load_configuration(device_file_name, device_options_file_name,
                             testbeds_file_name, gdm_config_file_name,
                             log_directory, adb_path)

    # Register USR1 signal to get exception messages from exception_queue
    signal.signal(signal.SIGUSR1,
                  common_utils.MethodWeakRef(self._process_exceptions))
    atexit.register(common_utils.MethodWeakRef(self.close))

  def backup_configs(self):
    """Backs up existing configuration files to a timestamped directory.

    Raises:
      DeviceError: unable to overwrite config files.

    Notes:
      Backs up configuration files to 'backup'
    """
    timestamp = datetime.datetime.utcnow().strftime("%Y%m%d-%H%M%S")
    if not os.path.exists(config.BACKUP_PARENT_DIRECTORY) or not os.access(
        config.BACKUP_PARENT_DIRECTORY, os.X_OK):
      raise errors.DeviceError(
          "Device overwrite error. "
          "Directory {} does not exist or is not executable. "
          "Unable to overwrite configs".format(config.BACKUP_PARENT_DIRECTORY))
    self.backup_directory = os.path.join(config.BACKUP_PARENT_DIRECTORY,
                                         "backup-%s" % str(timestamp))
    logger.info("Moving config files to the backup directory " +
                self.backup_directory)
    if not os.path.exists(self.backup_directory):
      os.makedirs(self.backup_directory)

    shutil.copyfile(self.device_file_name,
                    os.path.join(self.backup_directory, "devices.json"))
    shutil.copyfile(self.device_options_file_name,
                    os.path.join(self.backup_directory, "device_options.json"))
    shutil.copyfile(self.testbeds_file_name,
                    os.path.join(self.backup_directory, "testbeds.json"))
    shutil.copyfile(self.gdm_config_file_name,
                    os.path.join(self.backup_directory, "gdm.json"))

  def close(self):
    """Stops logger and closes all devices."""
    self.close_open_devices()
    gdm_logger.flush_queue_messages()
    gdm_logger.silence_progress_messages()

    if hasattr(self, "gdm_log_handler") and self.gdm_log_handler:
      gdm_logger.remove_handler(self.gdm_log_handler)
      self.gdm_log_handler.close()
      del self.gdm_log_handler

    if hasattr(self, "_exception_queue"):
      del self._exception_queue
    if hasattr(self, "_exception_queue_manager"):
      self._exception_queue_manager.shutdown()
      del self._exception_queue_manager

  def close_open_devices(self):
    """Closes all open devices."""
    for device in list(self._open_devices.values()):
      device.close()

  def close_device(self, identifier):
    """Closes open device via identifier.

    Args:
      identifier (str): device identifier. Name, serial_number, etc
    """
    device_name = self._get_device_name(identifier, raise_error=True)
    if device_name not in self._open_devices:
      return
    else:
      self._open_devices[device_name].close()

  def create_device(self,
                    identifier,
                    new_alias=None,
                    log_file_name=None,
                    log_directory=None,
                    log_to_stdout=None,
                    skip_recover_device=False,
                    make_device_ready="on",
                    filters=None,
                    log_name_prefix="") -> custom_types.Device:
    """Returns created device object by identifier specified.

    Args:
      identifier (str): The identifier string to identify a single device.
        For simulators, the identifier can be the simulator device type
      new_alias (str): A string to replace device's alias kept in file.
      log_file_name (str): A string log file name to use for log results.
      log_directory (str): A directory path to use for storing log file.
      log_to_stdout (bool): Enable streaming of log results to stdout
        (DEPRECATED).
      skip_recover_device (bool): Don't recover device if it fails ready
        check.
      make_device_ready (str): "on", "check_only", "off". Toggles
        make_device_ready.
      filters (list): paths to custom Parser filter files or directories to
        use.
      log_name_prefix (str): string to prepend to log filename.

    Returns:
      The device found and created by the identifier specified.

    Raises:
      ValueError: If identifier specified does not match a known device or
                  device is not currently connected.
      DeviceError: Device not connected
    """
    logger.debug("In create_device")

    if identifier.endswith("sim"):
      return self.create_device_sim(
          device_type=identifier,
          log_file_name=log_file_name,
          log_directory=log_directory,
          skip_recover_device=skip_recover_device,
          make_device_ready="off",
          filters=filters,
          log_name_prefix=log_name_prefix)

    if log_to_stdout is not None:
      logger.warn(
          "DEPRECATION WARNING: Support for the log_to_stdout argument is "
          "ending soon. To continue seeing the same output, please set "
          "debug_level to logging.INFO and remove log_to_stdout")

    if log_file_name is not None:
      logger.warn(
          "DEPRECATION WARNING: Support for log_file_name argument is "
          "ending soon. Please start using log_name_prefix argument instead.")

    self._type_check("identifier", identifier)
    device_name = self._get_device_name(identifier, raise_error=True)
    if device_name in self._open_devices:
      raise errors.DeviceError(
          "Device {name} already created. Call manager.get_open_device('{name}')."
          .format(name=device_name))
    if new_alias is not None:
      self.set_prop(device_name, "alias", new_alias)
    device_config = self.get_device_configuration(device_name)
    self._update_device_config(device_config, skip_recover_device,
                               make_device_ready, log_name_prefix, filters)
    device_type = device_config["persistent"]["device_type"]
    if not log_directory:
      # sets the device log directory to manager's log_directory
      log_directory = self.log_directory

    logger.info("Creating {}".format(device_name))
    device_class = self.get_supported_device_class(device_type)
    track_device = device_type not in self.get_supported_auxiliary_device_types(
    )
    device_inst = self._get_device_class(device_class, device_config,
                                         log_file_name, log_directory,
                                         track_device)
    try:
      device_inst.make_device_ready(make_device_ready)

    except errors.DeviceError:
      # ensure connections are closed down.
      device_inst.close()
      raise
    return device_inst

  def create_device_sim(self,
                        device_type,
                        log_file_name=None,
                        log_directory=None,
                        skip_recover_device=False,
                        make_device_ready="off",
                        filters=None,
                        log_name_prefix="",
                        build_info_kwargs=None):
    """Returns created simulated object by device_type specified.

    Args:
      device_type (str): The device type of the simulator.
      log_file_name (str): A string log file name to use for log results.
      log_directory (str): A directory path to use for storing log file.
      skip_recover_device (bool): Don't recover device if it fails ready
        check.
      make_device_ready (str): "on", "check_only", "off". Toggles
        make_device_ready.
      filters (list): paths to custom Parser filter files or directories to
        use.
      log_name_prefix (str): string to prepend to log filename.
      build_info_kwargs (dict): build info args by name to pass to upgrade
        method.

    Returns:
      Object: The device found and created by the device_type specified.

    Raises:
      ValueError: If identifier specified does not match a known device_type
      DeviceError: Device not connected
    """
    logger.info("In create_device_sim")
    if not log_directory:
      # sets the device log directory to manager's log_directory
      log_directory = self.log_directory

    device_config = {}
    self._update_device_config(device_config, skip_recover_device,
                               make_device_ready, log_name_prefix, filters)

    supported_device_class = self.get_supported_device_class(device_type)
    device_class = self._get_device_sim_class(supported_device_class,
                                              device_config, log_file_name,
                                              log_directory, build_info_kwargs)
    return device_class

  def create_devices(self,
                     device_list=None,
                     device_type=None,
                     log_to_stdout=None,
                     category="gazoo",
                     make_device_ready="on",
                     log_name_prefix=""):
    """Returns list of created device objects from device_list or connected devices.

    Args:
      device_list (list): list of mobly configs.
      device_type (str): filter to just return device instances of list
        type.
      log_to_stdout (bool): Enable streaming of log results to stdout
        (DEPRECATED).
      category (str): 'gazoo', 'other' or 'all' to filter connected devices.
      make_device_ready (str): "on", "check_only", "off". Toggles
        make_device_ready.
      log_name_prefix (str): string to prepend to log filename.

    Returns:
      list: device instances successfully created.

    Raises:
      ValueError: If an identifier specified does not match a known device
      or device is not currently connected.
    """
    logger.debug("In create_devices")
    if log_to_stdout is not None:
      logger.warn(
          "DEPRECATION WARNING: Support for the log_to_stdout argument is "
          "ending soon. To continue seeing the same output, please set "
          "debug_level to logging.INFO and remove log_to_stdout")

    devices = []

    if device_list is None:
      device_list = self.get_connected_devices(category)

    alias = None
    identifier = None
    for args in device_list:
      if isinstance(args, dict):  # translating potential mobly arguments
        if "id" in args:
          identifier = args["id"]
        elif "name" in args:
          identifier = args["name"]
        if "label" in args:
          alias = args["label"]
        elif "alias" in args:
          alias = args["alias"]

      elif isinstance(args, str):
        identifier = args

      # check if this device is the right type:
      if device_type is None or device_type.lower() == self.get_device_prop(
          identifier, "device_type"):
        devices.append(
            self.create_device(
                identifier,
                alias,
                make_device_ready=make_device_ready,
                log_name_prefix=log_name_prefix))
    return devices

  def create_log_parser(self, log_filename, filter_list=None):
    """Creates a LogParser object given a specified device type and filter list.

    Args:
        log_filename (str): filename containing raw, log event data
        filter_list (list): List of files or directories containing JSON
          filter files.

    Returns:
        LogParser: object which creates an event file by parsing a log file
        of the device type specified using the provided filter list
    """
    parser = event_parser_default.EventParserDefault(
        filters=filter_list,
        event_file_path="unknown.txt",
        device_name="unknown")
    return LogParser(parser, log_filename)

  def create_switchboard(
      self,
      communication_address,
      communication_type,
      device_name="unknown",
      log_path=None,
      force_slow=False,
      event_parser=None,
      **kwargs):
    """Creates a switchboard instance.

    Args:
      communication_address (str): primary device address for communication.
        For example, "123.45.56.123", ADB serial number, or serial port path.
      communication_type (str): identifier for the type of communication.
      device_name (str): device identifier. Used in stdout.
      log_path (str): path to write GDM device logs to.
      force_slow (bool): send device input at human speed. Used for devices with
        input speed limitations.
      event_parser (Parser): parses log stream into events and saves them to
        event file.
      **kwargs (dict): additional kwargs to pass onto the communication setup.

    Returns:
      SwitchboardDefault: instance of SwitchboardDefault.

    Raises:
      SwitchboardCreationError: if communication type not recognized.
    """
    if communication_type == "GENERIC_PROCESS":
      # Backwards compatibility with VDM
      communication_type = "PtyProcessComms"
    if communication_type not in extensions.communication_types:
      raise errors.SwitchboardCreationError(
          device_name,
          "Communication type {!r} is not in supported types: {}".format(
              communication_type, extensions.communication_types.keys()))

    if not log_path:
      log_path = self.create_log_path(device_name)

    logger.info("{} logging to file {}", device_name, log_path)

    comm_type_class = extensions.communication_types[communication_type]

    method_args = inspect.getfullargspec(
        comm_type_class.__init__).args[1:]  # remove self
    bad_keys = set(kwargs.keys()) - set(method_args)
    if bad_keys:
      raise errors.SwitchboardCreationError(
          device_name,
          "Communication Type {} does not support args {}. Supported: {}"
          .format(communication_type, bad_keys, method_args))

    try:
      comm_inst = comm_type_class(communication_address, **kwargs)
      switchboard_kwargs = comm_inst.get_switchboard_kwargs()
      additional_kwargs = {
          "device_name": device_name,
          "log_path": log_path,
          "force_slow": force_slow,
          "parser": event_parser,
          "exception_queue": self._exception_queue,
          "max_log_size": self.max_log_size,
      }
      switchboard_kwargs.update(additional_kwargs)

      return switchboard.SwitchboardDefault(**switchboard_kwargs)
    except Exception as err:
      raise errors.SwitchboardCreationError(device_name, repr(err))

  def delete(self, device_name, save_changes=True):
    """Delete the device from config dict and file.

    Args:
        device_name (str): name, serial_number, alias, or adb_serial of the
          device.
        save_changes (bool): if True, updates the config files.

    Raises:
        DeviceError: Device not found.

    Returns:
        None: if save_changes is True.
        tuple[dict, dict]: if save_changes is False, returns the new device
        configs: (devices, device_options).
    """
    devices = copy.deepcopy(self.persistent_dict)
    device_options = copy.deepcopy(self.options_dict)
    other_devices = copy.deepcopy(self.other_persistent_dict)
    other_device_options = copy.deepcopy(self.other_options_dict)
    device_name_arg = device_name
    device_name = self._get_device_name(device_name, raise_error=True)
    if device_name in devices and device_name in device_options:
      del devices[device_name]
      del device_options[device_name]
    elif device_name in other_devices and device_name in other_device_options:
      del other_devices[device_name]
      del other_device_options[device_name]
    else:
      raise errors.DeviceError(
          "Unable to find device {}".format(device_name_arg))

    device_config, device_options_config = self._make_device_configs(
        devices, other_devices, device_options, other_device_options)
    if save_changes:  # save and reload the config.
      self._save_config_to_file(device_config, self.device_file_name)
      self._save_config_to_file(device_options_config,
                                self.device_options_file_name)
      self.reload_configuration()
      logger.info("Deleted {}".format(device_name_arg))
    else:
      return (device_config, device_options_config)

  def detect(self,
             force_overwrite=False,
             static_ips=None,
             log_directory=None,
             save_changes=True,
             device_configs=None):
    """Detect new devices not present in config files.

    Args:
       force_overwrite (bool): Erase the current configs completely and
         re-detect everything.
       static_ips (list): list of static ips to detect.
       log_directory (str): alternative location to store log from default.
       save_changes (bool): if True, updates the config files.
       device_configs (None or tuple[dict, dict]): device configs
         (persistent, options) to pass to the device detector. If None, uses
         the current Manager configs.

    Returns:
        None: if save_changes is True.
        tuple[dict, dict]: if save_changes is False, returns
            the new device configs: (devices, device_options).

    Note:
       Overwrite saves the files to a backup directory.
    """
    if device_configs is None:
      device_config, options_config = self._make_device_configs(
          self.persistent_dict, self.other_persistent_dict, self.options_dict,
          self.other_options_dict)
    else:
      device_config, options_config = device_configs

    if not static_ips:
      static_ips = []
    elif isinstance(static_ips, str):
      static_ips = [ip_addr for ip_addr in static_ips.split(",") if ip_addr]
    if not log_directory:
      log_directory = self.log_directory

    if force_overwrite:
      comm_ports = [
          a_dict["persistent"].get("console_port_name", "")
          for name, a_dict in self._devices.items()
      ]
      static_ips += [
          comm_port for comm_port in comm_ports
          if host_utils.is_static_ip(comm_port)
      ]
      if save_changes:
        self.overwrite_configs()
      device_config, options_config = self._make_device_configs({}, {}, {}, {})

    detector = device_detector.DeviceDetector(
        manager=self,
        log_directory=log_directory,
        persistent_configs=device_config,
        options_configs=options_config,
        supported_auxiliary_device_classes=self
        .get_supported_auxiliary_device_classes())

    new_device_config, new_options_config = detector.detect_all_new_devices(
        static_ips)
    if save_changes:
      self._save_config_to_file(new_device_config, self.device_file_name)
      self._save_config_to_file(new_options_config,
                                self.device_options_file_name)
      self.reload_configuration()
      self.devices()
    else:
      return (new_device_config, new_options_config)

  def devices(self):
    """Prints a summary of device info.
    """

    self._print_device_info_by_category("gazoo")
    self._print_device_info_by_category("other")

    logger.info("{} total Gazoo device(s) available.".format(
        len(self.get_connected_devices())))

  def download_keys(self):
    """Downloads all required GDM keys if they don't exist locally."""
    for key_info in extensions.keys:
      host_utils.verify_key(key_info)

  @classmethod
  def get_all_supported_capabilities(cls):
    """Returns a map of all capability names supported by GDM.

    Returns:
        dict: map from capability name (str) to capability interface name
        (str).
              Example: {"file_transfer": "filetransferbase"}.
    """
    return copy.copy(extensions.capabilities)

  @classmethod
  def get_all_supported_capability_interfaces(cls):
    """Returns a map of all capability interface classes supported by GDM.

    Returns:
        dict: map from interface name (str) to capability interface class
        (type).
              Example: {"filetransferbase": <class FileTransferBase>}.
    """
    return copy.copy(extensions.capability_interfaces)

  @classmethod
  def get_all_supported_capability_flavors(cls):
    """Returns a map of all capability flavor classes supported by GDM.

    Returns:
        dict: map from flavor name (str) to capability flavor class (type).
              Example: {"filetransferscp": <class FileTransferScp>}.
    """
    return copy.copy(extensions.capability_flavors)

  @classmethod
  def get_all_supported_device_classes(cls):
    """Returns a list of all supported primary, sim, and auxiliary devices.

    Returns:
      list: All supported device types. Returns just categories asked for if
      requested.
    """
    all_classes = copy.copy(extensions.auxiliary_devices)
    all_classes += copy.copy(extensions.primary_devices)
    all_classes += copy.copy(extensions.virtual_devices)
    return all_classes

  def get_connected_devices(self, category="gazoo"):
    """Retrieve a list of connected devices for the category specified.

    Args:
      category (str): device category ('gazoo', 'other', or 'all') to
        retrieve.

    Returns:
      list: List of known connected devices.

    Note:
      If category is not specified then a list of all devices will be
      returned.
    """
    devices = self.get_devices(category)
    connected_devices = []
    for name in devices:
      if self.is_device_connected(name, category):
        connected_devices.append(name)
    return connected_devices

  def get_device_configuration(self, identifier, category="all"):
    """Returns the configuration for the device.

    Args:
        identifier (str): Name or alias to search for.
        category (str): device category ('gazoo', 'other' or 'all') to
          retrieve.

    Returns:
      dict: Configuration obtained for the device found.

    Raises:
      DeviceError: If identifier does not unique identify the device.

    Note:
      If category is not specified then all devices will be used to find
      the matching identifier.
    """
    # returns the device configuration
    device_name = self._get_device_name(identifier, category, raise_error=True)
    return self._get_device_configuration(device_name, category)

  def get_open_device_names(self):
    """Returns a list of open device names.

    Returns:
      list: open device names
    """
    return list(self._open_devices.keys())

  def get_open_device(self, identifier):
    """Returns device object if device is open.

    Args:
      identifier (str): device name, serial_number etc.

    Returns:
      GazooDeviceBase: device object

    Raises:
      DeviceError: if device not currently open
    """
    device_name = self._get_device_name(identifier, raise_error=True)
    if device_name not in self._open_devices:
      raise errors.DeviceError(
          "Device {} is not currently open".format(identifier))
    else:
      return self._open_devices[device_name]

  def get_open_devices(self):
    """Returns list of device objects."""
    return list(self._open_devices.values())

  def get_device_prop(self, device_name, prop=None):
    """Gets an prop's value for device or GDM configuration depends on identifier.

    Args:
        device_name (str): "manager", name, serial_number, alias, or
          adb_serial of the device.
        prop (str): Public prop available in device_options.json or gdm.json.
          Default is None.

    Returns:
      dict: device properties dicts if prop is None
      value: value of valid prop
    """

    if self._is_manager_config(device_name):
      return self._get_config_prop(prop)
    else:
      return self._get_device_prop(device_name, prop)

  @classmethod
  def get_supported_auxiliary_device_classes(cls):
    return copy.copy(extensions.auxiliary_devices)

  @classmethod
  def get_supported_auxiliary_device_types(cls):
    return [
        a_cls.DEVICE_TYPE
        for a_cls in cls.get_supported_auxiliary_device_classes()
    ]

  @classmethod
  def get_supported_device_capabilities(cls, device_type):
    """Returns a list of names of capabilities supported by the device type.

    This is a wrapper around GazooDeviceBase.get_supported_capabilities() to
    allow specifying device_type as a string.

    Args:
        device_type (str): device type to query for supported capabilities.

    Returns:
        list: list of capability names supported by this device type.
              For example, (["file_transfer", "usb_hub"]).
    """
    device_class = cls.get_supported_device_class(device_type)
    return device_class.get_supported_capabilities()

  @classmethod
  def get_supported_device_capability_flavors(cls, device_type):
    """Returns a set of all capability flavor classes supported by the device type.

    This is a wrapper around GazooDeviceBase.get_supported_capability_flavors()
    to allow specifying device_type as a string.

    Args:
      device_type (str): device type to query for supported capability flavors.

    Returns:
      set: capability flavor classes supported by this device type.
      Example: {<class 'DevicePowerDefault'>, <class 'FileTransferScp'>}.
    """
    device_class = cls.get_supported_device_class(device_type)
    return device_class.get_supported_capability_flavors()

  @classmethod
  def get_supported_device_class(cls, device_type):
    """Converts device type to device class.

    Args:
      device_type (str): device type.

    Returns:
       class: GazooDeviceBase-based class.

    Raises:
      DeviceError: if unknown type.
    """
    classes = [
        device_class for device_class in cls.get_all_supported_device_classes()
        if device_class.DEVICE_TYPE == device_type
    ]
    if classes:
      return classes[0]
    else:
      close_matches = difflib.get_close_matches(
          device_type, cls.get_supported_device_types())
      raise errors.DeviceError(
          "Device type {} is not known. Close matches: {}".format(
              device_type, ", ".join(close_matches)))

  @classmethod
  def get_supported_device_types(cls):
    """Returns a list of all supported device types.

    Returns:
      list: All supported device types.
    """
    return [
        a_cls.DEVICE_TYPE for a_cls in cls.get_all_supported_device_classes()
    ]

  @classmethod
  def get_supported_primary_device_classes(cls):
    return copy.copy(extensions.primary_devices)

  @classmethod
  def get_supported_primary_device_types(cls):
    return [
        a_cls.DEVICE_TYPE
        for a_cls in cls.get_supported_primary_device_classes()
    ]

  @classmethod
  def get_supported_virtual_device_classes(cls):
    return copy.copy(extensions.virtual_devices)

  @classmethod
  def get_supported_virtual_device_types(cls):
    return [
        a_cls.DEVICE_TYPE
        for a_cls in cls.get_supported_virtual_device_classes()
    ]

  def is_device_connected(self, identifier, category="all"):
    """Determine if device match identifier provided is connected for the category specified.

    Args:
      identifier (str): Name or alias to search for.
      category (str): device category ('gazoo', 'other', or 'all') to
        retrieve.

    Returns:
      bool: True if the matching devices is connected. False otherwise.

    Raises:
      ValueError: Identifier does not unique identify the device.

    Note:
      If category is not specified then the list of all devices will be used
      to find the matching identifier.
    """
    device_name = self._get_device_name(identifier, category, raise_error=True)

    device_config = self._get_device_configuration(device_name, category)
    device_type = device_config["persistent"]["device_type"].lower()
    try:
      device_class = self.get_supported_device_class(device_type)
      return device_class.is_connected(device_config)
    # Device configs might have devices listed that aren't supported
    except errors.DeviceError as err:
      logger.debug(err)
      return False

  def overwrite_configs(self):
    """Overwrite device configs.

    Raises:
      DeviceError: Device overwrite error.

    Note:
      Backs up existing configuration files and writes empty config files.
      Then reloads manager().
    """
    self.backup_configs()

    with open(self.device_file_name, "w") as open_file:
      json.dump({"devices": {}, "other_devices": {}}, open_file)
    with open(self.device_options_file_name, "w") as open_file:
      json.dump({"device_options": {}, "other_device_options": {}}, open_file)
    with open(self.testbeds_file_name, "w") as open_file:
      json.dump({"testbeds": {}}, open_file)
    with open(self.gdm_config_file_name, "w") as open_file:
      json.dump({}, open_file)

    self.reload_configuration()

  def get_usb_hub_props(self, device_identifier):
    """Dictionary of usb_hub information.

    Args:
        device_identifier (str): the device identifier.

    Returns:
        list: a dictionary of usb_hub props.

    Note: A usb hub will return an empty dictionary.
    Use port_map to see the devices connected to it.
    """
    props_dict = self.get_device_configuration(device_identifier)["persistent"]
    return {key: value for key, value in props_dict.items() if "usb" in key}

  def port_map(self):
    """Prints the USB Port Map.
    """
    usb_port_map = UsbPortMap(self)
    usb_port_map.print_port_map()

  def issue_devices(self,
                    devices,
                    method_name,
                    timeout=parallel_utils.TIMEOUT_PROCESS,
                    **kwargs):
    """Execute a device method in parallel for multiple devices.

    Args:
        devices (list): list of device identifiers.
        method_name (str): name of device method to execute in parallel.
        timeout (int): maximum amount of seconds to allow parallel methods
          to complete.
        **kwargs (dict): arguments to pass to device method.

    Returns:
        list: list of results from parallel calls.
    """
    if isinstance(devices, str):
      devices = devices.split(",")

    return self._issue_devices(devices, method_name, timeout, **kwargs)

  def issue_devices_all(self,
                        method_name,
                        timeout=parallel_utils.TIMEOUT_PROCESS,
                        **kwargs):
    """Execute a device method in parallel for all connected devices.

    Args:
        method_name (str): name of device method to execute in parallel.
        timeout (int): maximum amount of seconds to allow parallel methods
          to complete.
        **kwargs (dict): arguments to pass to device method.

    Returns:
        list: list of results from parallel calls.

    Raises:
        DeviceError: if no devices are connected.
    """
    devices = self.get_connected_devices()
    if not devices:
      raise errors.DeviceError("No devices are connected.")

    return self._issue_devices(devices, method_name, timeout, **kwargs)

  def issue_devices_match(self,
                          match,
                          method_name,
                          timeout=parallel_utils.TIMEOUT_PROCESS,
                          **kwargs):
    """Execute a device method in parallel for connected devices that match a given string.

    Args:
      match (str): wildcard-supported string to match against device
        names, i.e. "raspberrypi*" will call provided method on all connected
        Raspberry Pis.
      method_name (str): name of device method to execute in parallel.
      timeout (int): maximum amount of seconds to allow parallel methods
        to complete.
      **kwargs (dict): arguments to pass to device method.

    Returns:
      list: results from parallel calls.

    Raises:
      DeviceError: if provided wildcard does not match any connected devices.
    """
    devices = fnmatch.filter(self.get_connected_devices(), match)
    if not devices:
      raise errors.DeviceError('No devices match "{}".'.format(match))

    return self._issue_devices(devices, method_name, timeout, **kwargs)

  def redetect(self, device_name, log_directory=None):
    """Delete a device from the device configuration and then do a detect to find it again.

    Args:
        device_name (str): identifier for device.
        log_directory (str): alternative location to store log from default.

    Raises:
        DeviceError: if device is not found.
    """
    device_name = self._get_device_name(device_name, raise_error=True)
    static_ips = None
    try:
      hub_name, hub_port = self._get_device_usb_hub_name_and_port(device_name)
      logger.debug("{} before redetect, found device_usb_hub_name: {}, "
                   "device_usb_port: {}".format(device_name, hub_name,
                                                hub_port))
      comms_port = self.get_device_configuration(
          device_name)["persistent"]["console_port_name"]
      if host_utils.is_static_ip(comms_port):
        static_ips = [comms_port]
    except errors.DeviceError as err:
      logger.info(err)
    usb_hub = None
    original_power_mode = None
    if hub_name and hub_port:
      usb_hub = self.create_device(hub_name)
      original_power_mode = usb_hub.switch_power.get_mode(hub_port)
      if original_power_mode != "sync":
        usb_hub.switch_power.set_mode("sync", hub_port)

    device_configs_after_delete = self.delete(device_name, save_changes=False)
    new_device_config, new_options_config = self.detect(
        static_ips=static_ips,
        log_directory=log_directory,
        save_changes=False,
        device_configs=device_configs_after_delete)
    if usb_hub and original_power_mode != "sync":
      usb_hub.switch_power.set_mode(original_power_mode, hub_port)

    if (device_name not in new_device_config["devices"] and
        device_name not in new_device_config["other_devices"]):
      # Either the device changed its name (if it was physically replaced),
      # or detection of that device failed.
      raise errors.DeviceError(
          "Device {device_name} was not successfully redetected. "
          "Keeping the old device configs. "
          "If you replaced the device, use 'gdm delete {device_name}' "
          "and then 'detect' instead of 'redetect'.".format(
              device_name=device_name))

    # Preserve optional properties set by users
    if device_name in new_options_config["device_options"]:
      new_options_config["device_options"][device_name] = (
          self.options_dict[device_name])
    if device_name in new_options_config["other_device_options"]:
      new_options_config["other_device_options"][device_name] = (
          self.other_options_dict[device_name])

    self._save_config_to_file(new_device_config, self.device_file_name)
    self._save_config_to_file(new_options_config, self.device_options_file_name)
    self.reload_configuration()
    logger.info("Re-detected {}".format(device_name))

  def reload_configuration(self,
                           device_file_name=None,
                           options_file_name=None,
                           testbeds_file_name=None,
                           gdm_config_file_name=None,
                           log_directory=None,
                           adb_path=None):
    """Used when a config file or setting is changed.

    Args:
      device_file_name (str): Device file name.
      options_file_name (str): Device options file name.
      testbeds_file_name (str): Testbeds file name.
      gdm_config_file_name (str): GDM config file name.
      log_directory (str): Log directory.
      adb_path (str): adb path.
    """
    if device_file_name is None:
      device_file_name = self.device_file_name
    if options_file_name is None:
      options_file_name = self.device_options_file_name
    if testbeds_file_name is None:
      testbeds_file_name = self.testbeds_file_name
    if gdm_config_file_name is None:
      gdm_config_file_name = self.gdm_config_file_name

    self._load_configuration(device_file_name, options_file_name,
                             testbeds_file_name, gdm_config_file_name,
                             log_directory, adb_path)

  def _add_correct_value_to_config(self, key, value, default):
    """Add new attribute to self.config dict.

    Args:
      key (str): key's name. if there is no such key, a new key will be
        created.
      value (str): key's corresponding value. If value is not None, the
        value is used.
      default (str): default value for key

    Raises:
      DeviceError: if set attribute is None.

    Notes:
      The value is a arbitrary string and there is no check for such string.
      i.e. This function won't check if the value is a existing path.
    """
    if value is not None:
      self.config[key] = value

    elif key not in self.config:
      self.config[key] = default

    final_value = self.config[key]
    if final_value is None:
      raise errors.DeviceError(
          "Final value for {} should not be None".format(key))

    setattr(self, key, final_value)

  def _add_correct_path_to_config(self, key, path, default):
    """Add new path configuration to self.config dict.

    Args:
      key (str): key's name. if there is no such key, a new key will be
        created.
      path (str): key's corresponding value. If value is not None, this
        value is used.
      default (str): default value for key.

    Raises:
      DeviceError: if final value is none or directory doesn't exist.

    Notes:
      The path is a string represent path, and the function check if the
      path exist or not.
      If there exist such path, the key and value are added into dict.
    """
    if path is not None:
      self.config[key] = path
    elif key not in self.config:
      self.config[key] = default

    final_value = self.config[key]
    if final_value is None:
      raise errors.DeviceError(
          "Final value for {} should not be None".format(key))
    if "/" in final_value and not os.path.isdir(os.path.dirname(final_value)):
      raise errors.DeviceError(
          "Parent directory of {} for {} doesn't exist".format(
              final_value, key))
    if ".json" not in final_value and not os.path.exists(final_value):
      raise errors.DeviceError("File {} for {} doesn't exist".format(
          final_value, key))
    setattr(self, key, final_value)

  def _combine_devices_and_device_options(self, devices_dict,
                                          device_options_dict):
    """Loads the "persistent" items and "optional" items from json file.

    Args:
      devices_dict (dict): Contains "persistent" items of device.
      device_options_dict (dict): Contains "optional" items of device.

    Returns:
      tuple: (aliases, devices, persistent, options)

    Notes:
      This is called once when the manager is initialized. It iterates thru the
      devices.json dictionary, loading the "persistent" items from there, and
      the "optional" items from the  device_options.json file.

      When it is finished: self._devices is a tree-shaped dict.  Each key is the
      name of a device, and the "value" is another dict with "persistent" and
      "optional" sub-dicts). self.persistent_dict has JUST the persistent info
      (subset of the info in self._devices). self.options_dict has JUST the
      optional info (subset of the info in self._devices). TODO: Perhaps
      refactor to remove self.persistent_dict and self.options_dict and just use
      self._devices?
    """
    aliases = self._create_aliases(devices_dict, device_options_dict)
    devices = {}
    persistent = {}
    options = {}
    for name in devices_dict:
      persistent_dict = devices_dict[name]
      persistent_dict["name"] = name
      options_dict = device_options_dict[name]
      new_device_config = {
          "persistent": persistent_dict,
          "options": options_dict
      }
      devices[name] = new_device_config
      persistent[name] = persistent_dict
      options[name] = options_dict
    return aliases, devices, persistent, options

  def _create_aliases(self, devices_dict, devices_options_dict):
    """Creates a dictionary of aliases of Devices.

    Args:
      devices_dict (dict): Contains devices info.
      devices_options_dict (dict): Contains devices options info.

    Returns:
      dict: aliases with keys(name, adb_serial, serial_number, alias from
      device options).

    Notes:
      Maps the lowercase versions of the name, adb_serial, serial_number to
      the device's name.
    """
    aliases = {}
    alias_props = [
        "serial_number", "console_port_name", "alias", "adb_serial",
        "hub_port_name", "secondary_communication_address", "ip_address"
    ]
    for name in devices_dict:
      aliases[name.lower()] = name
      all_props = dict(devices_dict[name], **devices_options_dict[name])

      for prop in alias_props:
        alias = all_props.get(prop, None)
        if alias:
          aliases[alias.lower()] = name
    return aliases

  def create_log_path(self, device_name, name_prefix=""):
    """Returns the full path of log filename using the information provided.

    Args:
        device_name (str): to use in the log filename
        name_prefix (str): string to prepend to the start of the log file.

    Returns:
        str: Path to log filename using the information provided.
    """
    log_timestamp = time.strftime("%Y%m%d-%H%M%S")
    if name_prefix:
      log_file_name = "{}-{}-{}.txt".format(name_prefix, device_name,
                                            log_timestamp)
    else:
      log_file_name = "{}-{}.txt".format(device_name, log_timestamp)
    return os.path.join(self.log_directory, log_file_name)

  @classmethod
  def device_has_capabilities(cls, device_type, capability_names):
    """Check whether a device type supports all of the given capabilities.

    This is a wrapper around GazooDeviceBase.has_capabilities() to allow
        specifying device_type as a string.

    Args:
        device_type (str): device type to query for supported capabilities.
        capability_names (list): list of capability names.
            Capability names are strings. They can be:
                - capability names ("file_transfer"),
                - capability interface names ("filetransferbase"),
                - capability flavor names ("filetransferscp").
            If an interface name or capability name is specified, the behavior
            is identical: any capability flavor which implements the given
            interface will match. If a flavor name is specified, only that
            capability flavor will match. Different kinds of capability names
            can be used together (["usb_hub", "filetransferscp"]).

    Returns:
        bool: True if all capabilities are supported by the device type,
        False otherwise.
    """
    device_class = cls.get_supported_device_class(device_type)
    return device_class.has_capabilities(capability_names)

  def _get_aliases(self, category):
    """Returns a dict of all device name aliases for the category specified.

    Args:
      category (str): Indicates the device category ('gazoo', 'other', or
        'all') to retrieve.

    Returns:
      dict: Device name aliases for the category.
    """
    aliases = {}
    aliases.update(self.aliases if category in ["all", "gazoo"] else {})
    aliases.update(self.other_aliases if category in ["all", "other"] else {})
    return aliases

  def _get_config_prop(
      self, prop: Optional[str] = None
  ) -> Union[custom_types.PropertyValue,
             Dict[str, Dict[str, custom_types.PropertyValue]]]:
    """Returns the value of an GDM config property.

    Args:
      prop: Manager property available in gdm.json.

    Returns:
      GDM config property value.

    Raises:
      DeviceError: Property is not present in the config file.
    """
    if prop:
      if prop not in self.config:
        raise errors.DeviceError(
            "Unable to find prop {} in manager config".format(prop))
      return self.config[prop]
    else:  # return dict of all props
      return {"settable": self.config.copy()}

  def _get_device_configuration(self, name, category):
    """Returns the configuration for the device.

    Args:
      name (str): Key to use for obtaining device configuration.
      category (str): Indicates the device category ('gazoo', 'other' or
        'all') to retrieve.

    Returns:
      dict: Configuration obtained for the device found
      None: if device not found.

    Notes:
      If category is not specified then all devices will be used to find the
      matching identifier.
    """
    devices = self.get_devices(category)
    if name in devices:
      return devices[name]
    else:
      logger.debug("Unable to find device {}", name)
      return None

  def _get_device_name(self, identifier, category="all", raise_error=False):
    """Returns the device key name for the device.

    Args:
        identifier (str): Name or alias to search for.
        category (str): Device category. Options: ('gazoo', 'other', or
          'all') Default: 'all'.
        raise_error (bool): raise error if unable to find device. Default:
          False

    Returns:
      str: Device key name to use for the identifier specified.
      None: If device not found and raise_error is false.

    Raises:
      DeviceError: If identifier is not string/unicode or device does not
      exist

    Notes:
      If category is not specified then all devices will be used to find
      the matching identifier.
    """
    if not isinstance(identifier, str):
      raise errors.DeviceError("Device identifier '{}' should be a string. "
                               "but instead it is a {}".format(
                                   str(identifier), str(type(identifier))))
    aliases = self._get_aliases(category)
    identifier = identifier.lower()
    if identifier not in aliases:
      logger.debug("Unable to find device {}", identifier)
      if not raise_error:
        return None
      close_matches = difflib.get_close_matches(identifier, aliases)
      raise errors.DeviceError(
          "Device {} is not known. Close matches: {}".format(
              identifier, ", ".join(close_matches)))
    return aliases[identifier]

  def _get_device_usb_hub_name_and_port(self, device_name):
    """Returns the hub_name and port for the USB hub configured for the device.

    Args:
        device_name (str): device identifier.

    Returns:
        tuple: usb_hub_name, usb_port These will be set to None if the
        device or the
                                      property are not defined.
    """
    hub_name = None
    hub_port = None
    if device_name in self.options_dict:
      hub_name = self.options_dict[device_name].get("usb_hub", None)
      hub_port = self.options_dict[device_name].get("usb_port", None)
    if device_name in self.persistent_dict:
      hub_name = self.persistent_dict[device_name].get("device_usb_hub_name",
                                                       hub_name)
      hub_port = self.persistent_dict[device_name].get("device_usb_port",
                                                       hub_port)
    if hub_port:
      hub_port = int(hub_port)
    return hub_name, hub_port

  def get_devices(self, category):
    """Returns a dict of all devices for the category specified.

    Args:
        category (str): device category (gazoo, other, or all) to retrieve

    Returns:
        dict: device_name, device_instance
    """
    devices = {}
    devices.update(self._devices if category in ["all", "gazoo"] else {})
    devices.update(self.other_devices if category in ["all", "other"] else {})
    return devices

  def _get_device_prop(self, identifier, prop=None):
    """Gets a prop's value for a device if the device and the prop exist.

    Args:
      identifier (str): name, serial_number, alias, or adb_serial of the
        device.
      prop (str): Name of a single property to fetch.

    Returns:
      dict: if prop is None.  The full device_config
      value: The value of that prop if prop specified.

    Raises:
      DeviceError: if device doesn't exist or not queryable
      ValueError: if property not available

    Note: The full device config is a multi-level dict.  Top level keys are
        "persistent", "optional", "dynamic", with prop_value pairs in the
        next level. If device is connected it will fill out dynamic properties.
    """
    # Validate the prop input.
    self._type_check("Prop", prop, allowed_types=(type(None), str))

    logger.debug("Getting prop for identifier: {} Attr: {}".format(
        identifier, prop))
    close_device = True
    device_name = self._get_device_name(identifier, raise_error=True)
    if device_name in self.get_open_device_names():
      close_device = False
      device = self.get_open_device(device_name)
    else:
      device = self.create_device(identifier, make_device_ready="off")
    try:
      if prop:  # Return a specific property.
        prop = prop.lower()
        if prop in ["communication_type", "device_type"]:
          prop = prop.upper()
        # Don't need to activate device:
        offline_props = list(device.get_persistent_property_names())
        offline_props += list(device.get_optional_property_names())
        offline_props += list(device.props["optional"].keys())
        if prop in offline_props:
          return device.get_property(prop)
        device.make_device_ready()
        return device.get_property(prop, raise_error=True)

      # prop is None, so return the whole list of props
      device_config = {}
      device_config["persistent"] = device.get_persistent_properties()
      device_config["optional"] = device.get_optional_properties()
      if device.connected:
        try:
          device.make_device_ready()
          device_config["dynamic"] = device.get_dynamic_properties()
        except errors.DeviceError:
          device_config["dynamic"] = {"connected": True, "status": "Unhealthy."}

      else:
        device_config["dynamic"] = {"connected": False}
      return device_config
    finally:
      if close_device:
        device.close()

  def _get_attributes_list(self, device_config, just_props=False):
    """Return public attributes of class.

    Args:
      device_config (dict): device configuration dict
      just_props (bool): If just_props, returns non-method attributes.

    Returns:
      list: attributes
    """
    device_type = device_config["persistent"]["device_type"]
    this_class = self.get_supported_device_class(device_type)
    all_attributes = [a for a in dir(this_class) if not a.startswith("_")]
    if just_props:
      all_props = [
          a for a in all_attributes
          if not inspect.isroutine(hasattr(this_class, a))
      ]
      return all_props
    return all_attributes

  def _get_device_class(self,
                        device_class,
                        device_config,
                        log_file_name,
                        log_directory,
                        track_device=True):
    """Returns the device class after adding it to the list of shared resources."""
    device = device_class(
        self,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    if track_device:
      self._open_devices[device.name] = device
    return device

  def _get_device_sim_class(self, device_class, device_config, log_file_name,
                            log_directory, build_info_kwargs):
    """Returns the device class after adding it to the list of shared resources."""
    device = device_class(
        self,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory,
        build_info_kwargs=build_info_kwargs)
    self._open_devices[device.name] = device
    return device

  def _issue_devices(self,
                     devices,
                     method_name,
                     timeout=parallel_utils.TIMEOUT_PROCESS,
                     **kwargs):
    """Execute a device method in parallel for multiple devices.

    Args:
        devices (list): list of device identifiers.
        method_name (str): name of device method to execute in parallel.
        timeout (int): maximum amount of seconds to allow parallel methods
          to complete.
        **kwargs (dict): arguments to pass to device method.

    Returns:
        list: list of results from parallel calls, if any.

    Raises:
        DeviceError: if a provided device does not have a method with name
        method_name.
    """
    device_names = []
    parameter_dicts = {}
    for device_id in devices:

      # collect names of devices that support method
      device_name = self._get_device_name(device_id, raise_error=True)
      device_type = self.get_device_configuration(
          device_name)["persistent"]["device_type"]
      device_class = self.get_supported_device_class(device_type)
      if not hasattr(device_class, method_name):
        raise errors.DeviceError("Device {} does not support method {}".format(
            device_id, method_name))
      else:
        device_names.append(device_name)
        parameter_dicts[device_type] = kwargs

    # create device instances
    device_instances = [
        self.create_device(device_name, make_device_ready="off")
        for device_name in device_names
    ]

    # execute device methods in parallel using parallel_utils
    try:
      results = parallel_utils.issue_devices_parallel(method_name,
                                                      device_instances,
                                                      parameter_dicts, timeout)
    finally:
      for device in device_instances:
        device.close()

    if results:
      return results

  def _load_configuration(self,
                          device_file_name=None,
                          device_options_file_name=None,
                          testbeds_file_name=None,
                          gdm_config_file_name=None,
                          log_directory=None,
                          adb_path=None):
    """Loads GDM configuration.

    Args:
        device_file_name (str): Device file name.
        device_options_file_name (str): Device options file name.
        testbeds_file_name (str): Testbeds file name.
        gdm_config_file_name (str): GDM config file name.
        log_directory (str): Log directory.
        adb_path (str): "adb" binary path.

    Raises:
        DeviceError: failed to load Manager config.
    """
    self.gdm_config_file_name = gdm_config_file_name

    # create and configure self.config from gdm.conf
    self._load_gdm_configuration()

    # configure self.config from parameters
    self._add_correct_path_to_config("device_file_name", device_file_name,
                                     config.DEFAULT_DEVICE_FILE)

    self._add_correct_path_to_config("device_options_file_name",
                                     device_options_file_name,
                                     config.DEFAULT_OPTIONS_FILE)

    self._add_correct_path_to_config("testbeds_file_name", testbeds_file_name,
                                     config.DEFAULT_TESTBEDS_FILE)

    self._add_correct_path_to_config("log_directory", log_directory,
                                     config.DEFAULT_LOG_DIRECTORY)
    self._add_correct_value_to_config(config.ADB_BIN_PATH_CONFIG, adb_path, "")

    self._load_devices()
    self._load_other_devices()
    self._load_testbeds()

  def _load_devices(self):
    devices = self._load_config(self.device_file_name, config.DEVICES_KEYS[0])
    device_options = self._load_config(self.device_options_file_name,
                                       config.OPTIONS_KEYS[0])
    (self.aliases, self._devices, self.persistent_dict,
     self.options_dict) = self._combine_devices_and_device_options(
         devices, device_options)

  def _load_other_devices(self):
    other_devices = self._load_config(self.device_file_name,
                                      config.DEVICES_KEYS[1])
    other_device_options = self._load_config(self.device_options_file_name,
                                             config.OPTIONS_KEYS[1])
    (self.other_aliases, self.other_devices, self.other_persistent_dict,
     self.other_options_dict) = self._combine_devices_and_device_options(
         other_devices, other_device_options)

  def _load_testbeds(self):
    self.testbeds = self._load_config(self.testbeds_file_name,
                                      config.TESTBED_KEYS[0])

  def _load_gdm_configuration(self):
    self._generate_config_files()
    self.config = self._load_config(self.gdm_config_file_name, None)

  def _generate_config_files(self):
    """Generate config files if files or folders are missing.

    Raises:
      DeviceError: if unable to access files.
    """
    # Ensure all folders exist and have correct permissions.
    expected_permissions = "755"
    for folder in config.REQUIRED_FOLDERS:
      if not os.path.exists(folder):
        os.makedirs(folder)
      permissions = oct(os.stat(folder).st_mode)[-3:]
      if permissions != expected_permissions:
        try:
          os.chmod(folder, int(expected_permissions, 8))
        except OSError:
          raise errors.DeviceError(
              f"Unable to set correct permissions on {folder} without "
              f"sudo. Current permissions: {permissions}. Please run "
              f"sudo chmod -R {expected_permissions} {folder}'")

    # Ensure all files exist and are correctly populated
    config_info = {
        "device_file_name": config.DEFAULT_DEVICE_FILE,
        "device_options_file_name": config.DEFAULT_OPTIONS_FILE,
        "testbeds_file_name": config.DEFAULT_TESTBEDS_FILE,
        "log_directory": config.DEFAULT_LOG_DIRECTORY,
        "cli_extension_packages": [],
    }
    self._create_dict_if_doesnt_exist(config_info,
                                      config.DEFAULT_GDM_CONFIG_FILE)
    self._create_dict_if_doesnt_exist({key: {} for key in config.DEVICES_KEYS},
                                      config.DEFAULT_DEVICE_FILE)
    self._create_dict_if_doesnt_exist({key: {} for key in config.OPTIONS_KEYS},
                                      config.DEFAULT_OPTIONS_FILE)
    self._create_dict_if_doesnt_exist({key: {} for key in config.TESTBED_KEYS},
                                      config.DEFAULT_TESTBEDS_FILE)

  def _create_dict_if_doesnt_exist(self, a_dict, file_path):
    """Populates a file if it doesn't yet exist."""
    if not os.path.exists(file_path):
      with open(file_path, "w+") as open_file:
        json.dump(a_dict, open_file)

  def _load_config(self, file_name, key=None):
    """Loads a json config from a file into a dict and returns the dict.

    Args:
      file_name (str): The json file path.
      key (str): Indicates which dict entry caller is interested in.

    Returns:
      dict: The return value depends on key value. If the key value is None,
            the whole dict will return to caller.
      str: If there's a key specified, returns just the entry from that key.

    Raises:
      DeviceError: Device load configuration failed.
    """
    if not os.path.exists(file_name):
      raise errors.DeviceError(
          "Device load configuration failed. "
          "File {} is not found. \n Current directory: {}".format(
              file_name, os.getcwd()))

    with open(file_name, "r") as open_file:
      for i in range(2):
        try:
          conf = json.load(open_file)
          break
        except Exception as err:
          msg = "Unable to parse GDM config file as a json file. {!r}".format(
              err)

          if i == 1:
            raise errors.DeviceError(msg)
          else:
            logger.debug(msg)

    if key is None:
      return conf
    if key not in conf:
      return {}
    else:
      return conf[key]

  def _make_device_configs(self, devices, other_devices, device_options,
                           other_device_options):
    """Creates device configs with the provided dictionaries."""
    device_config = {"devices": devices, "other_devices": other_devices}
    device_options_config = {
        "device_options": device_options,
        "other_device_options": other_device_options
    }
    return (device_config, device_options_config)

  def set_prop(self, device_name, prop, value):
    """Sets a property's value for device or GDM configuration depends on identifier.

    Args:
      device_name (str): "manager", name, serial_number, alias, or
        adb_serial of the device.
      prop (str): Public prop available in device_options.json or gdm.json.
      value (str): Input value for specific property.

    Returns:
      bool: True if set_prop operation success.
    """
    if self._is_manager_config(device_name):
      return self._set_config_prop(prop, value)
    else:
      return self._set_device_prop(device_name, prop, value)

  def remove_prop(self, identifier, prop):
    """Removes a property for device if it's an optional property.

    Args:
      identifier (str): device identifier.
      prop (str): name of device property

    Raises:
      DeviceError: Attempts to remove property from manager or non-optional
      property.
    """
    if self._is_manager_config(identifier):
      raise errors.DeviceError("Not configured for manager.")
    else:
      self._remove_device_prop(identifier, prop)

  def _is_manager_config(self, identifier):
    """Check if identifier belongs to manager or not.

    Args:
      identifier (str): To check if it belongs to manager configuration.

    Returns:
      bool: If identifier is "manager" return True, otherwise return False.
    """
    if identifier == "manager":
      return True
    else:
      return False

  def _print_device_info_by_category(self, category="gazoo"):
    """Logs the device info in a special format.

    Args:
      category (str): 'gazoo' or 'other'.
    """
    format_line = "{:26} {:15} {:20} {:20} {:10}"
    if category == "gazoo":
      device_dict = self._devices
      title = "Device"
      connected_title = "Connected"
      good_status = "connected"
    else:
      device_dict = self.other_devices
      title = "Other Devices"
      connected_title = "Available"
      good_status = "available"
    logger.info(
        format_line.format(title, "Alias", "Type", "Model", connected_title))
    logger.info(
        format_line.format("-" * 26, "-" * 15, "-" * 20, "-" * 20, "-" * 10))
    for name in sorted(device_dict.keys()):
      device_config = device_dict[name]
      device_type = device_config["persistent"]["device_type"]
      model = device_config["persistent"]["model"]
      alias = device_config["options"].get("alias",
                                           "<undefined>") or u"<undefined>"
      if self.is_device_connected(name, category):
        status = good_status
      else:
        status = "unavailable"

      logger.info(format_line.format(name, alias, device_type, model, status))
    logger.info("")

  def _process_exceptions(self, signum, frame):  # pylint: disable=unused-argument
    """Retrieves and raises exceptions in exception_queue from all subprocesses created.

    Args:
        signum (int): signal number provided to this signal handler
        frame (object): current stack frame provided to this signal handler

    Raises:
        DeviceError: containing exception information received
    """

    try:
      exception_message = self._exception_queue.get_nowait()
      self._exception_queue.task_done()
    except (queue.Empty, ValueError):
      exception_message = "Exception queue missing exception message on SIGUSR1"
    except AttributeError:
      exception_message = "Exception queue deleted by parent process"

    raise errors.DeviceError(exception_message)

  def _set_device_prop(self, identifier, prop, value):
    """Sets a property's value for device.

    Args:
      identifier (str): name, serial_number, alias, or adb_serial of the
        device.
      prop (str): Public prop available in device_options.json.
      value (str): Input value for specific property.

    Raises:
      DeviceError: Device not found.

    Returns:
      bool: True if _set_device_prop operation success.
    """
    self._type_check("Property name", prop)
    self._type_check(prop, value, allowed_types=(str, type(None), int))
    device_name = self._get_device_name(identifier, raise_error=True)
    close_device = True
    if device_name in self.get_open_device_names():
      close_device = False
      device = self.get_open_device(device_name)
    else:
      device = self.create_device(identifier, make_device_ready="off")
    try:
      if prop == "alias":
        self._realign_alias(value, device.alias, device.name)
      device.set_property(prop, value)
    finally:
      if close_device:
        device.close()
    if device_name in self._devices:
      a_dict = self._devices
    else:
      a_dict = self.other_devices
    a_dict[device_name]["options"][prop] = value
    self._save_config_to_file(
        {
            "device_options": self.options_dict,
            "other_device_options": self.other_options_dict
        }, self.device_options_file_name)
    return True

  def _remove_device_prop(self, identifier, prop):
    """Removes prop from device config dict and file if in 'options'.

    Args:
      identifier (str): name, serial_number, alias, or adb_serial of the
        device.
      prop (str): Public prop available in device_options.json.

    Raises:
      DeviceError: Device not found or property not optional.
    """
    device_config = self.get_device_configuration(identifier)
    if prop in device_config["options"]:
      del device_config["options"][prop]
      self._save_config_to_file(
          {
              "device_options": self.options_dict,
              "other_device_options": self.other_options_dict
          }, self.device_options_file_name)
    else:
      raise errors.DeviceError(
          "Property {} is not an optional property for {}.".format(
              prop, identifier))

  def _realign_alias(self, new_alias, old_alias, name):
    """Modifying self.aliases when device's alias changes.

    Args:
      new_alias (str): new alias for device.
      old_alias (str): old alias of device.
      name (str): device name.
    """
    if old_alias is not None:
      del self.aliases[old_alias.lower()]
    if new_alias is not None:
      self.aliases[new_alias.lower()] = name

  def _set_config_prop(self,
                       prop: str,
                       value: custom_types.PropertyValue) -> None:
    """Sets an GDM config property.

    Args:
      prop: Manager property available in gdm.json.
      value: Property value.
    """
    self._type_check("value", value, allowed_types=(str, type(None), int, list))
    self._type_check("prop", prop)
    self.config[prop] = value
    # save property to json file
    self._save_config_to_file(self.config, self.gdm_config_file_name)

  def _save_config_to_file(self, a_dict, file_path):
    """Saves the dictionary to the given file."""
    logger.debug("Overwriting {}", file_path)
    config_directory = os.path.dirname(file_path)
    temp_file_path = os.path.join(config_directory,
                                  "temp_config_{}.json".format(os.getpid()))
    with open(temp_file_path, "w") as open_file:
      json.dump(a_dict, open_file, sort_keys=True, indent=4)
    shutil.move(temp_file_path, file_path)

  def _type_check(self, name, value, allowed_types=(str,)):
    """Sanity checking of (string or None) input values.

    Args:
      name (str): Name of the item being checked (for error messages).
      value (str): The item being verified.
      allowed_types (tuple): Tuple of "types".

    Raises:
      DeviceError: if value doesn't pass type check.

    Notes:
      Verifies the type, and for string inputs confirms there are no spaces,
      tabs, non-ascii, carriage-returns.
    """
    if not isinstance(value, allowed_types):
      raise errors.DeviceError(
          "{} {}'s type is not one of {}, but a {!r} instead".format(
              name, value, [t.__name__ for t in allowed_types],
              type(value).__name__))
    if isinstance(value, str):
      if not value:
        raise errors.DeviceError("{} {}'s value is illegally empty".format(
            name, value))
      if len(value.encode("ascii", "ignore")) != len(value):
        raise errors.DeviceError(
            "{} {} contains illegal non-ascii characters.".format(name, value))

      if " " in value:
        raise errors.DeviceError("{} {} contains illegal spaces".format(
            name, value))
      if "\t" in value:
        raise errors.DeviceError("{} {} contains illegal tabs".format(
            name, value))
      if "\n" in value:
        raise errors.DeviceError("{} {} contains illegal returns".format(
            name, value))

  def _update_device_config(self, device_config, skip_recover_device,
                            make_device_ready, log_name_prefix, filters):
    """Updates the device config based on the input params."""
    device_config["skip_recover_device"] = skip_recover_device
    device_config["make_device_ready"] = make_device_ready
    device_config["log_name_prefix"] = log_name_prefix
    device_config["filters"] = filters
    if "persistent" and "options" in device_config:
      if "device_usb_hub_name" not in device_config["persistent"]:
        if "usb_hub" in device_config["options"] and device_config["options"][
            "usb_hub"]:
          device_config["persistent"]["device_usb_hub_name"] = \
              device_config["options"]["usb_hub"]
      if "device_usb_port" not in device_config["persistent"]:
        if "usb_port" in device_config["options"] and device_config["options"][
            "usb_port"]:
          device_config["persistent"]["device_usb_port"] = \
              device_config["options"]["usb_port"]

  def __del__(self):
    self.close()
